#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Methods for eNRTL activity coefficient method.

Only applicable to liquid/electrolyte phases

Many thanks to C.-C. Chen for his assistance and suggestions on testing and
verifying the model.

Reference:

Song, Y. and Chen, C.-C., Symmetric Electrolyte Nonrandom Two-Liquid Activity
Coefficient Model, Ind. Eng. Chem. Res., 2009, Vol. 48, pgs. 7788–7797

Note that "charge number" in the paper referes to the absolute value of the
ionic charge.
"""
from pyomo.environ import Expression, exp, log, Set, units as pyunits

from .ideal import Ideal
from .enrtl_reference_states import Symmetric
from .enrtl_parameters import ConstantAlpha, ConstantTau
from idaes.generic_models.properties.core.generic.utility import (
    get_method, get_component_object as cobj)
from idaes.generic_models.properties.core.generic.generic_property import \
    StateIndex
from idaes.core.util.constants import Constants
from idaes.core.util.exceptions import BurntToast, ConfigurationError
import idaes.logger as idaeslog


# Set up logger
_log = idaeslog.getLogger(__name__)


DefaultAlphaRule = ConstantAlpha
DefaultTauRule = ConstantTau
DefaultRefState = Symmetric

# Closest appraoch parameter - implemented as a global constant for now
# This is not something the user should be changing in most cases
ClosestApproach = 14.9


class ENRTL(Ideal):
    # Add attribute indicating support for electrolyte systems
    electrolyte_support = True

    @staticmethod
    def build_parameters(b):
        # Build additional indexing sets
        pblock = b.parent_block()
        ion_pair = []
        for i in pblock.cation_set:
            for j in pblock.anion_set:
                ion_pair.append(i+", "+j)
        b.ion_pair_set = Set(initialize=ion_pair)

        comps = pblock.solvent_set | pblock.solute_set | b.ion_pair_set
        comp_pairs = []
        comp_pairs_sym = []
        for i in comps:
            for j in comps:
                if i in pblock.solvent_set | pblock.solute_set or i != j:
                    comp_pairs.append((i, j))
                    if (j, i) not in comp_pairs_sym:
                        comp_pairs_sym.append((i, j))
        b.component_pair_set = Set(initialize=comp_pairs)
        b.component_pair_set_symmetric = Set(initialize=comp_pairs_sym)

        # Check options for alpha rule
        if (b.config.equation_of_state_options is not None and
                "alpha_rule" in b.config.equation_of_state_options):
            b.config.equation_of_state_options[
                "alpha_rule"].build_parameters(b)
        else:
            DefaultAlphaRule.build_parameters(b)

        # Check options for tau rule
        if (b.config.equation_of_state_options is not None and
                "tau_rule" in b.config.equation_of_state_options):
            b.config.equation_of_state_options[
                "tau_rule"].build_parameters(b)
        else:
            DefaultTauRule.build_parameters(b)

    @staticmethod
    def common(b, pobj):
        pname = pobj.local_name

        molecular_set = b.params.solvent_set | b.params.solute_set

        # Check options for alpha rule
        if (pobj.config.equation_of_state_options is not None and
                "alpha_rule" in pobj.config.equation_of_state_options):
            alpha_rule = pobj.config.equation_of_state_options[
                "alpha_rule"].return_expression
        else:
            alpha_rule = DefaultAlphaRule.return_expression

        # Check options for tau rule
        if (pobj.config.equation_of_state_options is not None and
                "tau_rule" in pobj.config.equation_of_state_options):
            tau_rule = pobj.config.equation_of_state_options[
                "tau_rule"].return_expression
        else:
            tau_rule = DefaultTauRule.return_expression

        # Check options for reference state
        if (pobj.config.equation_of_state_options is not None and
                "reference_state" in pobj.config.equation_of_state_options):
            ref_state = pobj.config.equation_of_state_options[
                "reference_state"]
        else:
            ref_state = DefaultRefState

        # ---------------------------------------------------------------------
        # Calculate composition terms at both actual and reference state
        # Calculate reference state mole fractions
        ref_state.ref_state(b, pname)

        # Ionic Strength
        def rule_I(b):  # Eqn 62
            return (0.5*sum(b.mole_frac_phase_comp_true[pname, c] *
                            b.params.get_component(c).config.charge**2
                            for c in b.params.ion_set))
        b.add_component(pname+"_ionic_strength",
                        Expression(rule=rule_I,
                                   doc="Ionic strength"))

        def rule_I_ref(b):  # Eqn 62 evaluated at reference state
            x = getattr(b, pname+"_x_ref")
            return (0.5*sum(x[c] *
                            b.params.get_component(c).config.charge**2
                            for c in b.params.ion_set))
        b.add_component(pname+"_ionic_strength_ref",
                        Expression(rule=rule_I_ref,
                                   doc="Ionic strength at reference state"))

        # Calculate mixing factors
        def rule_X(b, j):  # Eqn 21
            if (pname, j) not in b.params.true_phase_component_set:
                return Expression.Skip
            elif j in b.params.cation_set or j in b.params.anion_set:
                return (b.mole_frac_phase_comp_true[pname, j] *
                        abs(cobj(b, j).config.charge))
            else:
                return b.mole_frac_phase_comp_true[pname, j]

        b.add_component(pname+"_X",
                        Expression(b.params.true_species_set,
                                   rule=rule_X,
                                   doc="Charge x mole fraction term"))

        def rule_X_ref(b, j):  # Eqn 21 evaluated at reference state
            x = getattr(b, pname+"_x_ref")
            if (pname, j) not in b.params.true_phase_component_set:
                return Expression.Skip
            elif j in b.params.cation_set or j in b.params.anion_set:
                return (x[j]*abs(cobj(b, j).config.charge))
            else:
                return x[j]

        b.add_component(
            pname+"_X_ref",
            Expression(b.params.true_species_set,
                       rule=rule_X_ref,
                       doc="Charge x mole fraction term at reference state"))

        def rule_Y(b, j):
            if cobj(b, j).config.charge < 0:
                # Anion
                dom = b.params.anion_set
            else:
                dom = b.params.cation_set

            X = getattr(b, pname+"_X")
            return X[j]/sum(X[i] for i in dom)  # Eqns 36 and 37
        # Y is a charge ratio, and thus independent of x for symmetric state
        # TODO: This may need to change for the unsymmetric state
        b.add_component(pname+"_Y",
                        Expression(b.params.ion_set,
                                   rule=rule_Y,
                                   doc="Charge composition"))

        # ---------------------------------------------------------------------
        # Long-range terms
        # Average molar volume of solvent
        def rule_vol_mol_solvent(b):  # Eqn 77
            if len(b.params.solvent_set) == 1:
                s = b.params.solvent_set.first()
                return ENRTL.get_vol_mol_pure(b, "liq", s, b.temperature)
            else:
                return (sum(b.mole_frac_phase_comp_true[pname, s] *
                            ENRTL.get_vol_mol_pure(b, "liq", s, b.temperature)
                            for s in b.params.solvent_set) /
                        sum(b.mole_frac_phase_comp_true[pname, s]
                            for s in b.params.solvent_set))
        b.add_component(pname+"_vol_mol_solvent",
                        Expression(rule=rule_vol_mol_solvent,
                                   doc="Mean molar volume of solvent"))

        # Mean relative permitivity of solvent
        def rule_eps_solvent(b):  # Eqn 78
            if len(b.params.solvent_set) == 1:
                s = b.params.solvent_set.first()
                return get_method(b, "relative_permittivity_liq_comp", s)(
                            b, cobj(b, s), b.temperature)
            else:
                return (sum(b.mole_frac_phase_comp_true[pname, s] *
                            get_method(b, "relative_permittivity_liq_comp", s)(
                                b, cobj(b, s), b.temperature) *
                            b.params.get_component(s).mw
                            for s in b.params.solvent_set) /
                        sum(b.mole_frac_phase_comp_true[pname, s] *
                            b.params.get_component(s).mw
                            for s in b.params.solvent_set))
        b.add_component(pname+"_relative_permittivity_solvent",
                        Expression(
                            rule=rule_eps_solvent,
                            doc="Mean relative permittivity  of solvent"))

        # Debye-Huckel parameter
        def rule_A_DH(b):  # Eqn 61
            # Note: Where the paper refers to the dielectric constant, it
            # actually means the electric permittivity of the solvent
            # eps = eps_r*eps_0 (units F/m)
            # Note that paper is missing a required 4*pi term
            v = pyunits.convert(getattr(b, pname+"_vol_mol_solvent"),
                                pyunits.m**3/pyunits.mol)
            eps = getattr(b, pname+"_relative_permittivity_solvent")
            eps0 = Constants.vacuum_electric_permittivity
            return ((1/3)*(2*Constants.pi*Constants.avogadro_number/v)**0.5 *
                    (Constants.elemental_charge**2/(
                        4*Constants.pi *
                        eps*eps0*Constants.boltzmann_constant *
                        b.temperature))**(3/2)
                    )
        b.add_component(pname+"_A_DH",
                        Expression(
                            rule=rule_A_DH,
                            doc="Debye-Huckel parameter"))

        # Long-range (PDH) contribution to activity coefficient
        def rule_log_gamma_pdh(b, j):
            A = getattr(b, pname+"_A_DH")
            Ix = getattr(b, pname+"_ionic_strength")
            I0 = getattr(b, pname+"_ionic_strength_ref")
            rho = ClosestApproach
            if j in molecular_set:
                # Eqn 69
                # Note typo in original paper. Correct power for I is (3/2)
                return (2*A*Ix**(3/2)/(1+rho*Ix**(1/2)))
            elif j in b.params.ion_set:
                # Eqn 70
                z = abs(cobj(b, j).config.charge)
                return (-A*((2*z**2/rho) *
                            log((1+rho*Ix**0.5)/(1+rho*I0**0.5)) +
                            (z**2*Ix**0.5 - 2*Ix**(3/2)) / (1+rho*Ix**0.5) -
                            (2*Ix*I0**-0.5) / (1+rho*I0**0.5) *
                            ref_state.ndIdn(b, pname, j)))
            else:
                raise BurntToast(
                    "{} eNRTL model encountered unexpected component."
                    .format(b.name))
        b.add_component(
            pname+"_log_gamma_pdh",
            Expression(b.params.true_species_set,
                       rule=rule_log_gamma_pdh,
                       doc="Long-range contribution to activity coefficient"))

        # ---------------------------------------------------------------------
        # Local Contribution Terms
        # For the symmetric state, all of these are independent of composition
        # TODO: For the unsymmetric state, it may be necessary to recalculate
        # Calculate alphas for all true species pairings
        def rule_alpha_expr(b, i, j):
            Y = getattr(b, pname+"_Y")
            if ((pname, i) not in b.params.true_phase_component_set or
                    (pname, j) not in b.params.true_phase_component_set):
                return Expression.Skip
            elif ((i in molecular_set) and
                    (j in molecular_set)):
                # alpha equal user provided parameters
                return alpha_rule(b, pobj, i, j, b.temperature)
            elif (i in b.params.cation_set and j in molecular_set):
                # Eqn 32
                return sum(
                    Y[k] *
                    alpha_rule(
                        b, pobj, (i+", "+k), j, b.temperature)
                    for k in b.params.anion_set)
            elif (j in b.params.cation_set and i in molecular_set):
                # Eqn 32
                return sum(
                    Y[k] *
                    alpha_rule(
                        b, pobj, (j+", "+k), i, b.temperature)
                    for k in b.params.anion_set)
            elif (i in b.params.anion_set and j in molecular_set):
                # Eqn 33
                return sum(
                    Y[k] *
                    alpha_rule(
                        b, pobj, (k+", "+i), j, b.temperature)
                    for k in b.params.cation_set)
            elif (j in b.params.anion_set and i in molecular_set):
                # Eqn 33
                return sum(
                    Y[k] *
                    alpha_rule(
                        b, pobj, (k+", "+j), i, b.temperature)
                    for k in b.params.cation_set)
            elif (i in b.params.cation_set and j in b.params.anion_set):
                # Eqn 34
                if len(b.params.cation_set) > 1:
                    return sum(Y[k]*alpha_rule(
                        b, pobj, (i+", "+j), (k+", "+j), b.temperature)
                               for k in b.params.cation_set)
                else:
                    return 0.2
            elif (i in b.params.anion_set and j in b.params.cation_set):
                # Eqn 35
                if len(b.params.anion_set) > 1:
                    return sum(Y[k]*alpha_rule(
                        b, pobj, (j+", "+i), (j+", "+k), b.temperature)
                               for k in b.params.anion_set)
                else:
                    return 0.2
            elif ((i in b.params.cation_set and j in b.params.cation_set) or
                  (i in b.params.anion_set and j in b.params.anion_set)):
                # No like-ion interactions
                return Expression.Skip
            else:
                raise BurntToast(
                    "{} eNRTL model encountered unexpected component pair {}."
                    .format(b.name, (i, j)))

        b.add_component(pname+"_alpha",
                        Expression(b.params.true_species_set,
                                   b.params.true_species_set,
                                   rule=rule_alpha_expr,
                                   doc="Non-randomness parameters"))

        # Calculate G terms
        def rule_G_expr(b, i, j):
            Y = getattr(b, pname+"_Y")

            def _G_appr(b, pobj, i, j, T):  # Eqn 23
                if i != j:
                    return exp(-alpha_rule(b, pobj, i, j, T) *
                               tau_rule(b, pobj, i, j, T))
                else:
                    return 1

            if ((pname, i) not in b.params.true_phase_component_set or
                    (pname, j) not in b.params.true_phase_component_set):
                return Expression.Skip
            elif ((i in molecular_set) and
                    (j in molecular_set)):
                # G comes directly from parameters
                return _G_appr(b, pobj, i, j, b.temperature)
            elif (i in b.params.cation_set and j in molecular_set):
                # Eqn 38
                return sum(
                    Y[k] * _G_appr(b, pobj, (i+", "+k), j,  b.temperature)
                    for k in b.params.anion_set)
            elif (i in molecular_set and j in b.params.cation_set):
                # Eqn 40
                return sum(
                    Y[k] * _G_appr(b, pobj, i, (j+", "+k), b.temperature)
                    for k in b.params.anion_set)
            elif (i in b.params.anion_set and j in molecular_set):
                # Eqn 39
                return sum(
                    Y[k] * _G_appr(b, pobj, (k+", "+i), j, b.temperature)
                    for k in b.params.cation_set)
            elif (i in molecular_set and j in b.params.anion_set):
                # Eqn 41
                return sum(
                    Y[k] * _G_appr(b, pobj, i, (k+", "+j), b.temperature)
                    for k in b.params.cation_set)
            elif (i in b.params.cation_set and j in b.params.anion_set):
                # Eqn 42
                if len(b.params.cation_set) > 1:
                    return sum(Y[k] * _G_appr(
                        b, pobj, (i+", "+j), (k+", "+j), b.temperature)
                        for k in b.params.cation_set)
                else:
                    # This term does not exist for single cation systems
                    # However, need a valid result to calculate tau
                    return 1
            elif (i in b.params.anion_set and j in b.params.cation_set):
                # Eqn 43
                if len(b.params.anion_set) > 1:
                    return sum(Y[k] * _G_appr(
                        b, pobj, (j+", "+i), (j+", "+k), b.temperature)
                        for k in b.params.anion_set)
                else:
                    # This term does not exist for single anion systems
                    # However, need a valid result to calculate tau
                    return 1
            elif ((i in b.params.cation_set and j in b.params.cation_set) or
                  (i in b.params.anion_set and j in b.params.anion_set)):
                # No like-ion interactions
                return Expression.Skip
            else:
                raise BurntToast(
                    "{} eNRTL model encountered unexpected component pair {}."
                    .format(b.name, (i, j)))
        b.add_component(pname+"_G",
                        Expression(b.params.true_species_set,
                                   b.params.true_species_set,
                                   rule=rule_G_expr,
                                   doc="Local interaction G term"))

        # Calculate tau terms
        def rule_tau_expr(b, i, j):
            if ((pname, i) not in b.params.true_phase_component_set or
                    (pname, j) not in b.params.true_phase_component_set):
                return Expression.Skip
            elif ((i in molecular_set) and
                    (j in molecular_set)):
                # tau equal to parameter
                return tau_rule(b, pobj, i, j, b.temperature)
            elif ((i in b.params.cation_set and j in b.params.cation_set) or
                  (i in b.params.anion_set and j in b.params.anion_set)):
                # No like-ion interactions
                return Expression.Skip
            else:
                alpha = getattr(b, pname+"_alpha")
                G = getattr(b, pname+"_G")
                # Eqn 44
                return -log(G[i, j])/alpha[i, j]

        b.add_component(pname+"_tau",
                        Expression(b.params.true_species_set,
                                   b.params.true_species_set,
                                   rule=rule_tau_expr,
                                   doc="Binary interaction energy parameters"))

        # Local contribution to activity coefficient
        def rule_log_gamma_lc_I(b, s):
            X = getattr(b, pname+"_X")
            G = getattr(b, pname+"_G")
            tau = getattr(b, pname+"_tau")

            return log_gamma_lc(b, pname, s, X, G, tau)

        b.add_component(pname+"_log_gamma_lc_I",
                        Expression(
                            b.params.true_species_set,
                            rule=rule_log_gamma_lc_I,
                            doc="Local contribution at actual state"))

        def rule_log_gamma_lc_I0(b, s):
            X = getattr(b, pname+"_X_ref")
            G = getattr(b, pname+"_G")
            tau = getattr(b, pname+"_tau")

            return log_gamma_lc(b, pname, s, X, G, tau)

        b.add_component(pname+"_log_gamma_lc_I0",
                        Expression(
                            b.params.ion_set,
                            rule=rule_log_gamma_lc_I0,
                            doc="Local contribution at reference state"))

        def rule_log_gamma_lc(b, s):
            log_gamma_lc_I = getattr(b, pname+"_log_gamma_lc_I")
            if s in molecular_set:
                return log_gamma_lc_I[s]
            else:
                log_gamma_lc_I0 = getattr(b, pname+"_log_gamma_lc_I0")
                return log_gamma_lc_I[s] - log_gamma_lc_I0[s]

        b.add_component(
            pname+"_log_gamma_lc",
            Expression(
                b.params.true_species_set,
                rule=rule_log_gamma_lc,
                doc="Local contribution contribution to activity coefficient"))

        # Overall log gamma
        def rule_log_gamma(b, j):
            pdh = getattr(b, pname+"_log_gamma_pdh")
            lc = getattr(b, pname+"_log_gamma_lc")
            return pdh[j] + lc[j]

        b.add_component(
            pname+"_log_gamma",
            Expression(
                b.params.true_species_set,
                rule=rule_log_gamma,
                doc="Log of activity coefficient"))

        # Activity coefficient of apparent species
        def rule_log_gamma_pm(b, j):
            cobj = b.params.get_component(j)

            if "dissociation_species" in cobj.config:
                dspec = cobj.config.dissociation_species

                n = 0
                d = 0
                for s in dspec:
                    dobj = b.params.get_component(s)
                    ln_g = getattr(b, pname+"_log_gamma")[s]
                    n += abs(dobj.config.charge)*ln_g
                    d += abs(dobj.config.charge)

                return n/d
            else:
                return getattr(b, pname+"_log_gamma")[j]

        b.add_component(
            pname+"_log_gamma_appr",
            Expression(
                b.params.apparent_species_set,
                rule=rule_log_gamma_pm,
                doc="Log of mean activity coefficient"))

    @staticmethod
    def calculate_scaling_factors(b, pobj):
        pass

    @staticmethod
    def act_phase_comp(b, p, j):
        return b.mole_frac_phase_comp[p, j]*b.act_coeff_phase_comp[p, j]

    @staticmethod
    def act_phase_comp_true(b, p, j):
        ln_gamma = getattr(b, p+"_log_gamma")
        return b.mole_frac_phase_comp_true[p, j]*exp(ln_gamma[j])

    @staticmethod
    def act_phase_comp_appr(b, p, j):
        ln_gamma = getattr(b, p+"_log_gamma_appr")
        return b.mole_frac_phase_comp_apparent[p, j]*exp(ln_gamma[j])

    @staticmethod
    def log_act_phase_comp(b, p, j):
        if b.params.config.state_components == StateIndex.true:
            ln_gamma = getattr(b, p+"_log_gamma")
        else:
            ln_gamma = getattr(b, p+"_log_gamma_appr")
        return log(b.mole_frac_phase_comp[p, j]) + ln_gamma[p, j]

    @staticmethod
    def log_act_phase_comp_true(b, p, j):
        ln_gamma = getattr(b, p+"_log_gamma")
        return log(b.mole_frac_phase_comp_true[p, j]) + ln_gamma[p, j]

    @staticmethod
    def log_act_phase_comp_appr(b, p, j):
        ln_gamma = getattr(b, p+"_log_gamma_appr")
        return log(b.mole_frac_phase_comp_apparent[p, j]) + ln_gamma[p, j]

    @staticmethod
    def act_coeff_phase_comp(b, p, j):
        if b.params.config.state_components == StateIndex.true:
            ln_gamma = getattr(b, p+"_log_gamma")
        else:
            ln_gamma = getattr(b, p+"_log_gamma_appr")
        return exp(ln_gamma[j])

    @staticmethod
    def act_coeff_phase_comp_true(b, p, j):
        ln_gamma = getattr(b, p+"_log_gamma")
        return exp(ln_gamma[j])

    @staticmethod
    def act_coeff_phase_comp_appr(b, p, j):
        ln_gamma = getattr(b, p+"_log_gamma_appr")
        return exp(ln_gamma[j])

    @staticmethod
    def pressure_osm_phase(b, p):
        return (-ENRTL.gas_constant(b)*b.temperature *
                log(sum(b.act_phase_comp[p, j]
                        for j in b.params.solvent_set)) /
                b.vol_mol_phase[p])

    @staticmethod
    def vol_mol_phase(b, p):
        # eNRTL model uses apparent species for calculating molar volume
        # TODO : Need something more rigorus to handle concentrated solutions
        v_expr = 0
        for j in b.params.apparent_species_set:
            v_comp = ENRTL.get_vol_mol_pure(b, "liq", j, b.temperature)
            v_expr += b.mole_frac_phase_comp_apparent[p, j]*v_comp

        return v_expr


def log_gamma_lc(b, pname, s, X, G, tau):
    # General function for calculating local contributions
    # The same method can be used for both actual state and reference state
    # by providing different X, G and tau expressions.

    # Indicies in expressions use same names as source paper
    # mp = m'
    molecular_set = b.params.solvent_set | b.params.solute_set
    aqu_species = b.params.true_species_set - b.params._non_aqueous_set

    if (pname, s) not in b.params.true_phase_component_set:
        # Non-aqueous component
        return Expression.Skip
    if s in b.params.cation_set:
        c = s
        Z = b.params.get_component(c).config.charge

        # Eqn 26
        return Z*(
            sum((X[m]*G[c, m] /
                 sum(X[i]*G[i, m] for i in aqu_species)) *
                (tau[c, m] -
                 (sum(X[i]*G[i, m]*tau[i, m] for i in aqu_species) /
                  sum(X[i]*G[i, m] for i in aqu_species)))
                for m in molecular_set) +
            sum(X[i]*G[i, c]*tau[i, c]
                for i in (aqu_species-b.params.cation_set)) /
            sum(X[i]*G[i, c]
                for i in (aqu_species-b.params.cation_set)) +
            sum((X[a]*G[c, a] /
                 sum(X[i]*G[i, a]
                     for i in (aqu_species-b.params.anion_set))) *
                (tau[c, a] -
                 sum(X[i]*G[i, a]*tau[i, a]
                     for i in (aqu_species-b.params.anion_set)) /
                 sum(X[i]*G[i, a]
                     for i in (aqu_species-b.params.anion_set)))
                for a in b.params.anion_set))
    elif s in b.params.anion_set:
        a = s
        Z = abs(b.params.get_component(a).config.charge)

        # Eqn 27
        return Z*(
            sum((X[m]*G[a, m] /
                 sum(X[i]*G[i, m] for i in aqu_species)) *
                (tau[a, m] -
                 (sum(X[i]*G[i, m]*tau[i, m] for i in aqu_species) /
                  sum(X[i]*G[i, m] for i in aqu_species)))
                for m in molecular_set) +
            sum(X[i]*G[i, a]*tau[i, a]
                for i in (aqu_species-b.params.anion_set)) /
            sum(X[i]*G[i, a]
                for i in (aqu_species-b.params.anion_set)) +
            sum((X[c]*G[a, c] /
                 sum(X[i]*G[i, c]
                     for i in (aqu_species-b.params.cation_set))) *
                (tau[a, c] -
                 sum(X[i]*G[i, c]*tau[i, c]
                     for i in (aqu_species-b.params.cation_set)) /
                 sum(X[i]*G[i, c]
                     for i in (aqu_species-b.params.cation_set)))
                for c in b.params.cation_set))
    else:
        m = s
        # Eqn 25
        return (sum(X[i]*G[i, m]*tau[i, m] for i in aqu_species) /
                sum(X[i]*G[i, m] for i in aqu_species) +
                sum((X[mp]*G[m, mp] /
                     sum(X[i]*G[i, mp] for i in aqu_species)) *
                    (tau[m, mp] -
                     (sum(X[i]*G[i, mp]*tau[i, mp]
                          for i in aqu_species) /
                      sum(X[i]*G[i, mp] for i in aqu_species)))
                    for mp in molecular_set) +
                sum((X[c]*G[m, c] /
                     sum(X[i]*G[i, c]
                         for i in (aqu_species-b.params.cation_set))) *
                    (tau[m, c] -
                     (sum(X[i]*G[i, c]*tau[i, c]
                          for i in (aqu_species-b.params.cation_set)) /
                      sum(X[i]*G[i, c]
                          for i in (aqu_species-b.params.cation_set))))
                    for c in b.params.cation_set) +
                sum((X[a]*G[m, a] /
                     sum(X[i]*G[i, a]
                         for i in (aqu_species-b.params.anion_set))) *
                    (tau[m, a] -
                     (sum(X[i]*G[i, a]*tau[i, a]
                          for i in (aqu_species-b.params.anion_set)) /
                      sum(X[i]*G[i, a]
                          for i in (aqu_species-b.params.anion_set))))
                    for a in b.params.anion_set)
                )



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Test for Caprese's module for NMPC.
"""

import pytest
from pyomo.environ import (Block, ConcreteModel,  Constraint, Expression,
                           Set, SolverFactory, Var, value, 
                           TransformationFactory, TerminationCondition)
from pyomo.network import Arc
from pyomo.common.collections import ComponentSet
from pyomo.dae.flatten import flatten_dae_components

from idaes.core import (FlowsheetBlock, MaterialBalanceType, EnergyBalanceType,
        MomentumBalanceType)
from idaes.core.util.model_statistics import (degrees_of_freedom, 
        activated_equalities_generator)
from idaes.apps.caprese.categorize import (
        categorize_dae_variables,
        categorize_dae_variables_and_constraints,
        )
from idaes.apps.caprese.common.config import VariableCategory as VC
import idaes.logger as idaeslog
from idaes.apps.caprese.examples.cstr_model import make_model

__author__ = "Robert Parker"


@pytest.mark.unit
def test_categorize_1():
    """
    This test categorizes the enzyme cstr "as intended."
    Volume is used as a measurement, and solvent flow rates are
    fixed, but otherwise equations are as expected.
    """
    model = make_model(horizon=1, ntfe=5, ntcp=2)
    time = model.fs.time

    init_input_list = [
            model.fs.mixer.S_inlet.flow_vol[0],
            model.fs.mixer.E_inlet.flow_vol[0],
            ]
    init_input_set = ComponentSet(init_input_list)

    init_deriv_list = [
            model.fs.cstr.control_volume.energy_accumulation[0, 'aq'],
            *list(model.fs.cstr.control_volume.material_accumulation[0, 'aq', :]),
            ]
    init_deriv_set = ComponentSet(init_deriv_list)

    init_diff_list = [
            model.fs.cstr.control_volume.energy_holdup[0, 'aq'],
            *list(model.fs.cstr.control_volume.material_holdup[0, 'aq', :]),
            ]
    init_diff_set = ComponentSet(init_diff_list)

    init_fixed_list = [
            model.fs.mixer.E_inlet.temperature[0],
            model.fs.mixer.S_inlet.temperature[0],
            *list(model.fs.mixer.E_inlet.conc_mol[0, :]),
            *list(model.fs.mixer.S_inlet.conc_mol[0, :]),
            model.fs.cstr.outlet.conc_mol[0, 'Solvent'],
            model.fs.mixer.outlet.conc_mol[0, 'Solvent'],
            ]
    init_fixed_set = ComponentSet(init_fixed_list)

    init_meas_list = [
            model.fs.cstr.control_volume.energy_holdup[0, 'aq'],
            model.fs.cstr.control_volume.volume[0],
            *list(model.fs.cstr.control_volume.material_holdup[0, 'aq', :]),
            ]
    init_meas_set = ComponentSet(init_meas_list)
    # Solvent holdup is not a measurement; we measure volume instead
    init_meas_set.remove(
            model.fs.cstr.control_volume.material_holdup[0, 'aq', 'Solvent'])

    init_alg_list = [
        model.fs.cstr.outlet.flow_vol[0],
        model.fs.cstr.outlet.temperature[0],
        model.fs.cstr.inlet.flow_vol[0],
        model.fs.cstr.inlet.temperature[0],
        model.fs.mixer.outlet.flow_vol[0],
        model.fs.mixer.outlet.temperature[0],
        model.fs.cstr.control_volume.volume[0],
        *list(model.fs.cstr.control_volume.properties_out[0].flow_mol_comp[:]),
        *list(model.fs.cstr.inlet.conc_mol[0, :]),
        *list(model.fs.cstr.control_volume.properties_in[0].flow_mol_comp[:]),
        *list(model.fs.cstr.control_volume.rate_reaction_generation[0, 'aq', :]),
        *list(model.fs.mixer.mixed_state[0].flow_mol_comp[:]),
        *list(model.fs.mixer.E_inlet_state[0].flow_mol_comp[:]),
        *list(model.fs.mixer.S_inlet_state[0].flow_mol_comp[:]),
        *list(model.fs.cstr.outlet.conc_mol[0, :]),
        *list(model.fs.mixer.outlet.conc_mol[0, :]),
        *list(model.fs.cstr.control_volume.reactions[0].reaction_coef[:]),
        *list(model.fs.cstr.control_volume.reactions[0].reaction_rate[:]),
        *list(model.fs.cstr.control_volume.rate_reaction_extent[0, :]),
        ]
    init_alg_set = ComponentSet(init_alg_list)
    # solvent outlet concentrations are not algebraic variables as
    # it is fixed.
    init_alg_set.remove(model.fs.cstr.outlet.conc_mol[0, 'Solvent'])
    init_alg_set.remove(model.fs.mixer.outlet.conc_mol[0, 'Solvent'])

    scalar_vars, dae_vars = flatten_dae_components(model, time, ctype=Var)
    category_dict = categorize_dae_variables(dae_vars, time, init_input_list)

    input_vars = category_dict[VC.INPUT]
    diff_vars = category_dict[VC.DIFFERENTIAL]
    deriv_vars = category_dict[VC.DERIVATIVE]
    fixed_vars = category_dict[VC.FIXED]
    alg_vars = category_dict[VC.ALGEBRAIC]
    meas_vars = category_dict[VC.MEASUREMENT]

    assert len(input_vars) == len(init_input_set)
    for v in input_vars:
        assert v[0] in init_input_set

    assert len(deriv_vars) == len(init_deriv_set)
    for v in deriv_vars:
        assert v[0] in init_deriv_set

    assert len(diff_vars) == len(init_deriv_set)
    for v in diff_vars:
        assert v[0] in init_diff_set

    assert len(fixed_vars) == len(init_fixed_set)
    for v in fixed_vars:
        assert v[0] in init_fixed_set

    assert len(alg_vars) == len(init_alg_set)
    for v in alg_vars:
        assert v[0] in init_alg_set

    assert len(meas_vars) == len(init_meas_set)
    for v in meas_vars:
        assert v[0] in init_meas_set

    assert len(scalar_vars) == 0


@pytest.mark.unit
def test_categorize_2():
    """
    In this instance, temperature is "measured" (used as an initial condition)
    instead of energy_holdup, conc[P] is measured instead of holdup[P],
    and accumulation[C] is measured instead of holdup[C].
    """
    model = make_model(horizon=1, ntfe=5, ntcp=2)
    time = model.fs.time
    t0 = time.first()

    material_holdup = model.fs.cstr.control_volume.material_holdup
    material_accumulation = model.fs.cstr.control_volume.material_accumulation
    energy_holdup = model.fs.cstr.control_volume.energy_holdup
    energy_accumulation = model.fs.cstr.control_volume.energy_accumulation
    conc_mol = model.fs.cstr.outlet.conc_mol

    # Specify temperature instead of energy holdup
    energy_holdup[t0, 'aq'].unfix()
    model.fs.cstr.outlet.temperature[t0].fix(300)

    # Specify C_P instead of holdup
    material_holdup[t0, 'aq', 'P'].unfix()
    conc_mol[t0, 'P'].fix(0.)

    # Specify accumulation of C instead of holdup
    # You might want to do this if you want to start at steady state,
    # but don't know the value of every variable at the steady state
    # you want to start at...
    material_holdup[t0, 'aq', 'C'].unfix()
    material_accumulation[t0, 'aq', 'C'].fix(0.)

    init_input_list = [
            model.fs.mixer.S_inlet.flow_vol[0],
            model.fs.mixer.E_inlet.flow_vol[0],
            ]
    init_input_set = ComponentSet(init_input_list)

    init_deriv_list = [
            model.fs.cstr.control_volume.energy_accumulation[0, 'aq'],
            *list(model.fs.cstr.control_volume.material_accumulation[0, 'aq', :]),
            ]
    init_deriv_set = ComponentSet(init_deriv_list)

    init_diff_list = [
            model.fs.cstr.control_volume.energy_holdup[0, 'aq'],
            *list(model.fs.cstr.control_volume.material_holdup[0, 'aq', :]),
            ]
    init_diff_set = ComponentSet(init_diff_list)

    init_fixed_list = [
            model.fs.mixer.E_inlet.temperature[0],
            model.fs.mixer.S_inlet.temperature[0],
            *list(model.fs.mixer.E_inlet.conc_mol[0, :]),
            *list(model.fs.mixer.S_inlet.conc_mol[0, :]),
            model.fs.cstr.outlet.conc_mol[0, 'Solvent'],
            model.fs.mixer.outlet.conc_mol[0, 'Solvent'],
            ]
    init_fixed_set = ComponentSet(init_fixed_list)

    init_meas_list = [
            model.fs.cstr.outlet.temperature[0],
            model.fs.cstr.control_volume.volume[0],
            model.fs.cstr.control_volume.material_holdup[0, 'aq', 'S'],
            model.fs.cstr.control_volume.material_holdup[0, 'aq', 'E'],
            model.fs.cstr.control_volume.material_holdup[0, 'aq', 'S'],
            model.fs.cstr.control_volume.material_accumulation[0, 'aq', 'C'],
            model.fs.cstr.outlet.conc_mol[0,'P'],
            ]
    init_meas_set = ComponentSet(init_meas_list)
    # No need to remove solvent holdup here as it was not added to this list.

    init_alg_list = [
        model.fs.cstr.outlet.flow_vol[0],
        model.fs.cstr.outlet.temperature[0],
        model.fs.cstr.inlet.flow_vol[0],
        model.fs.cstr.inlet.temperature[0],
        model.fs.mixer.outlet.flow_vol[0],
        model.fs.mixer.outlet.temperature[0],
        model.fs.cstr.control_volume.volume[0],
        *list(model.fs.cstr.control_volume.properties_out[0].flow_mol_comp[:]),
        *list(model.fs.cstr.inlet.conc_mol[0, :]),
        *list(model.fs.cstr.control_volume.properties_in[0].flow_mol_comp[:]),
        *list(model.fs.cstr.control_volume.rate_reaction_generation[0, 'aq', :]),
        *list(model.fs.mixer.mixed_state[0].flow_mol_comp[:]),
        *list(model.fs.mixer.E_inlet_state[0].flow_mol_comp[:]),
        *list(model.fs.mixer.S_inlet_state[0].flow_mol_comp[:]),
        *list(model.fs.cstr.outlet.conc_mol[0, :]),
        *list(model.fs.mixer.outlet.conc_mol[0, :]),
        *list(model.fs.cstr.control_volume.reactions[0].reaction_coef[:]),
        *list(model.fs.cstr.control_volume.reactions[0].reaction_rate[:]),
        *list(model.fs.cstr.control_volume.rate_reaction_extent[0, :]),
        ]
    init_alg_set = ComponentSet(init_alg_list)
    # solvent outlet concentrations are not algebraic variables as
    # it is fixed.
    init_alg_set.remove(model.fs.cstr.outlet.conc_mol[0, 'Solvent'])
    init_alg_set.remove(model.fs.mixer.outlet.conc_mol[0, 'Solvent'])

    scalar_vars, dae_vars = flatten_dae_components(model, time, ctype=Var)
    category_dict = categorize_dae_variables(dae_vars, time, init_input_list)

    input_vars = category_dict[VC.INPUT]
    diff_vars = category_dict[VC.DIFFERENTIAL]
    deriv_vars = category_dict[VC.DERIVATIVE]
    fixed_vars = category_dict[VC.FIXED]
    alg_vars = category_dict[VC.ALGEBRAIC]
    meas_vars = category_dict[VC.MEASUREMENT]

    assert len(input_vars) == len(init_input_set)
    for v in input_vars:
        assert v[0] in init_input_set

    assert len(deriv_vars) == len(init_deriv_set)
    for v in deriv_vars:
        assert v[0] in init_deriv_set

    assert len(diff_vars) == len(init_deriv_set)
    for v in diff_vars:
        assert v[0] in init_diff_set

    assert len(fixed_vars) == len(init_fixed_set)
    for v in fixed_vars:
        assert v[0] in init_fixed_set

    assert len(alg_vars) == len(init_alg_set)
    for v in alg_vars:
        assert v[0] in init_alg_set

    assert len(meas_vars) == len(init_meas_set)
    for v in meas_vars:
        assert v[0] in init_meas_set

    assert len(scalar_vars) == 0


@pytest.mark.unit
def test_categorize_3():
    """
    In this test we fix one of the differential variables.
    This is the case where somebody wants to run an isothermal
    CSTR.
    """
    model = make_model(horizon=1, ntfe=5, ntcp=2)
    time = model.fs.time

    CV = model.fs.cstr.control_volume
    CV.energy_holdup.fix(300)
    CV.energy_accumulation_disc_eq.deactivate()

    init_input_list = [
            model.fs.mixer.S_inlet.flow_vol[0],
            model.fs.mixer.E_inlet.flow_vol[0],
            ]
    init_input_set = ComponentSet(init_input_list)

    init_deriv_list = [
            *list(model.fs.cstr.control_volume.material_accumulation[0, 'aq', :]),
            ]
    init_deriv_set = ComponentSet(init_deriv_list)

    init_diff_list = [
            *list(model.fs.cstr.control_volume.material_holdup[0, 'aq', :]),
            ]
    init_diff_set = ComponentSet(init_diff_list)

    init_fixed_list = [
            # Energy holdup has been fixed
            model.fs.cstr.control_volume.energy_holdup[0, 'aq'],
            model.fs.mixer.E_inlet.temperature[0],
            model.fs.mixer.S_inlet.temperature[0],
            *list(model.fs.mixer.E_inlet.conc_mol[0, :]),
            *list(model.fs.mixer.S_inlet.conc_mol[0, :]),
            model.fs.cstr.outlet.conc_mol[0, 'Solvent'],
            model.fs.mixer.outlet.conc_mol[0, 'Solvent'],
            ]
    init_fixed_set = ComponentSet(init_fixed_list)

    init_meas_list = [
            model.fs.cstr.control_volume.volume[0],
            *list(model.fs.cstr.control_volume.material_holdup[0, 'aq', :]),
            ]
    init_meas_set = ComponentSet(init_meas_list)
    # Solvent holdup is not a measurement; we measure volume instead
    init_meas_set.remove(
            model.fs.cstr.control_volume.material_holdup[0, 'aq', 'Solvent'])

    init_alg_list = [
        # Since energy_holdup is fixed, energy_accumulation is not
        # considered a derivative. Instead it is algebraic.
        model.fs.cstr.control_volume.energy_accumulation[0, 'aq'],
        model.fs.cstr.outlet.flow_vol[0],
        model.fs.cstr.outlet.temperature[0],
        model.fs.cstr.inlet.flow_vol[0],
        model.fs.cstr.inlet.temperature[0],
        model.fs.mixer.outlet.flow_vol[0],
        model.fs.mixer.outlet.temperature[0],
        model.fs.cstr.control_volume.volume[0],
        *list(model.fs.cstr.control_volume.properties_out[0].flow_mol_comp[:]),
        *list(model.fs.cstr.inlet.conc_mol[0, :]),
        *list(model.fs.cstr.control_volume.properties_in[0].flow_mol_comp[:]),
        *list(model.fs.cstr.control_volume.rate_reaction_generation[0, 'aq', :]),
        *list(model.fs.mixer.mixed_state[0].flow_mol_comp[:]),
        *list(model.fs.mixer.E_inlet_state[0].flow_mol_comp[:]),
        *list(model.fs.mixer.S_inlet_state[0].flow_mol_comp[:]),
        *list(model.fs.cstr.outlet.conc_mol[0, :]),
        *list(model.fs.mixer.outlet.conc_mol[0, :]),
        *list(model.fs.cstr.control_volume.reactions[0].reaction_coef[:]),
        *list(model.fs.cstr.control_volume.reactions[0].reaction_rate[:]),
        *list(model.fs.cstr.control_volume.rate_reaction_extent[0, :]),
        ]
    init_alg_set = ComponentSet(init_alg_list)
    # solvent outlet concentrations are not algebraic variables as
    # it is fixed.
    init_alg_set.remove(model.fs.cstr.outlet.conc_mol[0, 'Solvent'])
    init_alg_set.remove(model.fs.mixer.outlet.conc_mol[0, 'Solvent'])

    scalar_vars, dae_vars = flatten_dae_components(model, time, ctype=Var)
    category_dict = categorize_dae_variables(dae_vars, time, init_input_list)

    input_vars = category_dict[VC.INPUT]
    diff_vars = category_dict[VC.DIFFERENTIAL]
    deriv_vars = category_dict[VC.DERIVATIVE]
    fixed_vars = category_dict[VC.FIXED]
    alg_vars = category_dict[VC.ALGEBRAIC]
    meas_vars = category_dict[VC.MEASUREMENT]

    assert len(input_vars) == len(init_input_set)
    for v in input_vars:
        assert v[0] in init_input_set

    assert len(deriv_vars) == len(init_deriv_set)
    for v in deriv_vars:
        assert v[0] in init_deriv_set

    assert len(diff_vars) == len(init_deriv_set)
    for v in diff_vars:
        assert v[0] in init_diff_set

    assert len(fixed_vars) == len(init_fixed_set)
    for v in fixed_vars:
        assert v[0] in init_fixed_set

    assert len(alg_vars) == len(init_alg_set)
    for v in alg_vars:
        assert v[0] in init_alg_set

    assert len(meas_vars) == len(init_meas_set)
    for v in meas_vars:
        assert v[0] in init_meas_set

    assert len(scalar_vars) == 0


@pytest.mark.unit
def test_categorize_4():
    """
    This tests categorization when a psuedo-steady state
    approximation is used. Energy accumulation and accumulation of E
    are fixed, the corresponding initial conditions unfixed, and
    the corresponding discretization equations deactivated.
    """
    model = make_model(horizon=1, ntfe=5, ntcp=2)
    time = model.fs.time

    CV = model.fs.cstr.control_volume
    CV.energy_accumulation[:, 'aq'].fix(0.)
    CV.material_accumulation[:, 'aq', 'E'].fix(0.)
    CV.energy_holdup[0, 'aq'].unfix()
    CV.material_holdup[0, 'aq', 'E'].unfix()
    CV.energy_accumulation_disc_eq.deactivate()
    CV.material_accumulation_disc_eq.deactivate()

    init_input_list = [
            model.fs.mixer.S_inlet.flow_vol[0],
            model.fs.mixer.E_inlet.flow_vol[0],
            ]
    init_input_set = ComponentSet(init_input_list)

    init_deriv_list = [
            CV.material_accumulation[0, 'aq', 'C'],
            CV.material_accumulation[0, 'aq', 'S'],
            CV.material_accumulation[0, 'aq', 'P'],
            CV.material_accumulation[0, 'aq', 'Solvent'],
            ]
    init_deriv_set = ComponentSet(init_deriv_list)

    init_diff_list = [
            CV.material_holdup[0, 'aq', 'C'],
            CV.material_holdup[0, 'aq', 'S'],
            CV.material_holdup[0, 'aq', 'P'],
            CV.material_holdup[0, 'aq', 'Solvent'],
            ]
    init_diff_set = ComponentSet(init_diff_list)

    init_fixed_list = [
            model.fs.cstr.control_volume.energy_accumulation[0, 'aq'],
            CV.material_accumulation[0, 'aq', 'E'],
            model.fs.mixer.E_inlet.temperature[0],
            model.fs.mixer.S_inlet.temperature[0],
            *list(model.fs.mixer.E_inlet.conc_mol[0, :]),
            *list(model.fs.mixer.S_inlet.conc_mol[0, :]),
            model.fs.cstr.outlet.conc_mol[0, 'Solvent'],
            model.fs.mixer.outlet.conc_mol[0, 'Solvent'],
            ]
    init_fixed_set = ComponentSet(init_fixed_list)

    init_meas_list = [
            model.fs.cstr.control_volume.volume[0],
            CV.material_holdup[0,'aq','P'],
            CV.material_holdup[0,'aq','C'],
            CV.material_holdup[0,'aq','S'],
            ]
    init_meas_set = ComponentSet(init_meas_list)

    init_alg_list = [
        model.fs.cstr.control_volume.energy_holdup[0, 'aq'],
        CV.material_holdup[0, 'aq', 'E'],
        model.fs.cstr.outlet.flow_vol[0],
        model.fs.cstr.outlet.temperature[0],
        model.fs.cstr.inlet.flow_vol[0],
        model.fs.cstr.inlet.temperature[0],
        model.fs.mixer.outlet.flow_vol[0],
        model.fs.mixer.outlet.temperature[0],
        model.fs.cstr.control_volume.volume[0],
        *list(model.fs.cstr.control_volume.properties_out[0].flow_mol_comp[:]),
        *list(model.fs.cstr.inlet.conc_mol[0, :]),
        *list(model.fs.cstr.control_volume.properties_in[0].flow_mol_comp[:]),
        *list(model.fs.cstr.control_volume.rate_reaction_generation[0, 'aq', :]),
        *list(model.fs.mixer.mixed_state[0].flow_mol_comp[:]),
        *list(model.fs.mixer.E_inlet_state[0].flow_mol_comp[:]),
        *list(model.fs.mixer.S_inlet_state[0].flow_mol_comp[:]),
        *list(model.fs.cstr.outlet.conc_mol[0, :]),
        *list(model.fs.mixer.outlet.conc_mol[0, :]),
        *list(model.fs.cstr.control_volume.reactions[0].reaction_coef[:]),
        *list(model.fs.cstr.control_volume.reactions[0].reaction_rate[:]),
        *list(model.fs.cstr.control_volume.rate_reaction_extent[0, :]),
        ]
    init_alg_set = ComponentSet(init_alg_list)
    # solvent outlet concentrations are not algebraic variables as
    # it is fixed.
    init_alg_set.remove(model.fs.cstr.outlet.conc_mol[0, 'Solvent'])
    init_alg_set.remove(model.fs.mixer.outlet.conc_mol[0, 'Solvent'])

    scalar_vars, dae_vars = flatten_dae_components(model, time, ctype=Var)
    category_dict = categorize_dae_variables(dae_vars, time, init_input_list)

    input_vars = category_dict[VC.INPUT]
    diff_vars = category_dict[VC.DIFFERENTIAL]
    deriv_vars = category_dict[VC.DERIVATIVE]
    fixed_vars = category_dict[VC.FIXED]
    alg_vars = category_dict[VC.ALGEBRAIC]
    meas_vars = category_dict[VC.MEASUREMENT]

    assert len(input_vars) == len(init_input_set)
    for v in input_vars:
        assert v[0] in init_input_set

    assert len(deriv_vars) == len(init_deriv_set)
    for v in deriv_vars:
        assert v[0] in init_deriv_set

    assert len(diff_vars) == len(init_deriv_set)
    for v in diff_vars:
        assert v[0] in init_diff_set

    assert len(fixed_vars) == len(init_fixed_set)
    for v in fixed_vars:
        assert v[0] in init_fixed_set

    assert len(alg_vars) == len(init_alg_set)
    for v in alg_vars:
        assert v[0] in init_alg_set

    assert len(meas_vars) == len(init_meas_set)
    for v in meas_vars:
        assert v[0] in init_meas_set

    assert len(scalar_vars) == 0


@pytest.mark.unit
def test_categorize_error():
    model = make_model(horizon=1, ntfe=5, ntcp=2)
    time = model.fs.time

    scalar_vars, dae_vars = flatten_dae_components(model, time, ctype=Var)

    # Add a dummy var to treat as an input.
    # This var is not in `dae_vars`, so it will not be located during
    # categorization, which should fail.
    model.dummy_var = Var(time)

    init_input_list = [
            model.fs.mixer.S_inlet.flow_vol[0],
            model.fs.mixer.E_inlet.flow_vol[0],
            model.dummy_var[0],
            ]

    with pytest.raises(RuntimeError, match=r'Not all inputs could be found'):
        category_dict = categorize_dae_variables(
                dae_vars,
                time,
                init_input_list,
                )

    # Re-run flattener. Now `dummy_var` should be included in `dae_vars`.
    scalar_vars, dae_vars = flatten_dae_components(model, time, ctype=Var)
    category_dict = categorize_dae_variables(
            dae_vars,
            time,
            init_input_list,
            )


if __name__ == '__main__':
    test_categorize_1()
    test_categorize_2()
    test_categorize_3()
    test_categorize_4()
    test_categorize_error()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
import sys, os
from unittest.mock import patch

sys.path.append(os.path.abspath('..'))  # current folder is ~/tests
from idaes.surrogate.pysmo.polynomial_regression import (
    PolynomialRegression, FeatureScaling
)
import numpy as np
import pandas as pd
import pytest


class TestFeatureScaling:
    test_data_1d = [[x] for x in range(10)]
    test_data_2d = [[x, (x + 1) ** 2] for x in range(10)]
    test_data_3d = [[x, x + 10, (x + 1) ** 2 + x + 10] for x in range(10)]
    test_data_3d_constant = [[x, 10, (x + 1) ** 2 + 10] for x in range(10)]

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_data_scaling_01(self, array_type):
        input_array = array_type(self.test_data_1d)

        output_1, output_2, output_3 = FeatureScaling.data_scaling(input_array)
        expected_output_3 = np.array([[9]])
        expected_output_2 = np.array([[0]])
        expected_output_1 = (input_array - expected_output_2) / (expected_output_3 - expected_output_2)
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, np.array(expected_output_1).reshape(10, 1))

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_data_scaling_02(self, array_type):
        input_array = array_type(self.test_data_2d)
        output_1, output_2, output_3 = FeatureScaling.data_scaling(input_array)
        expected_output_3 = np.array([[9, 100]])
        expected_output_2 = np.array([[0, 1]])
        expected_output_1 = (input_array - expected_output_2) / (expected_output_3 - expected_output_2)
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_data_scaling_03(self, array_type):
        input_array = array_type(self.test_data_3d)
        output_1, output_2, output_3 = FeatureScaling.data_scaling(input_array)
        expected_output_3 = np.array([[9, 19, 119]])
        expected_output_2 = np.array([[0, 10, 11]])
        expected_output_1 = (input_array - expected_output_2) / (expected_output_3 - expected_output_2)
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_data_scaling_04(self, array_type):
        input_array = array_type(self.test_data_3d_constant)
        output_1, output_2, output_3 = FeatureScaling.data_scaling(input_array)
        expected_output_3 = np.array([[9, 10, 110]])
        expected_output_2 = np.array([[0, 10, 11]])
        scale = expected_output_3 - expected_output_2
        scale[scale == 0.0] = 1.0
        expected_output_1 = (input_array - expected_output_2) / scale
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [list])
    def test_data_scaling_05(self, array_type):
        input_array = array_type(self.test_data_2d)
        with pytest.raises(TypeError):
            FeatureScaling.data_scaling(input_array)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_data_unscaling_01(self, array_type):
        input_array = array_type(self.test_data_1d)
        output_1, output_2, output_3 = FeatureScaling.data_scaling(input_array)
        output_1 = output_1.reshape(output_1.shape[0], )
        un_output_1 = FeatureScaling.data_unscaling(output_1, output_2, output_3)
        np.testing.assert_array_equal(un_output_1, input_array.reshape(10, 1))

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_data_unscaling_02(self, array_type):
        input_array = array_type(self.test_data_2d)
        output_1, output_2, output_3 = FeatureScaling.data_scaling(input_array)
        un_output_1 = FeatureScaling.data_unscaling(output_1, output_2, output_3)
        np.testing.assert_array_equal(un_output_1, input_array)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_data_unscaling_03(self, array_type):
        input_array = array_type(self.test_data_3d)
        output_1, output_2, output_3 = FeatureScaling.data_scaling(input_array)
        un_output_1 = FeatureScaling.data_unscaling(output_1, output_2, output_3)
        np.testing.assert_array_equal(un_output_1, input_array)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_data_unscaling_04(self, array_type):
        input_array = array_type(self.test_data_3d_constant)
        output_1, output_2, output_3 = FeatureScaling.data_scaling(input_array)
        un_output_1 = FeatureScaling.data_unscaling(output_1, output_2, output_3)
        np.testing.assert_array_equal(un_output_1, input_array)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_data_unscaling_05(self, array_type):
        input_array = array_type(self.test_data_2d)
        output_1, output_2, output_3 = FeatureScaling.data_scaling(input_array)

        min_array = np.array([[1]])
        max_array = np.array([[5]])
        with pytest.raises(IndexError):
            FeatureScaling.data_unscaling(output_1, min_array, max_array)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_data_unscaling_06(self, array_type):
        input_array = array_type(self.test_data_2d)
        output_1, output_2, output_3 = FeatureScaling.data_scaling(input_array)

        min_array = np.array([[1, 2, 3]])
        max_array = np.array([[5, 6, 7]])
        with pytest.raises(IndexError):
            FeatureScaling.data_unscaling(output_1, min_array, max_array)


class TestPolynomialRegression:
    y = np.array(
        [[i, j, ((i + 1) ** 2) + ((j + 1) ** 2)] for i in np.linspace(0, 10, 21) for j in np.linspace(0, 10, 21)])
    full_data = {'x1': y[:, 0], 'x2': y[:, 1], 'y': y[:, 2]}
    training_data = [[i, j, ((i + 1) ** 2) + ((j + 1) ** 2)] for i in np.linspace(0, 10, 5) for j in
                     np.linspace(0, 10, 5)]
    test_data = [[i, (i + 1) ** 2] for i in range(10)]
    test_data_large = [[i, (i + 1) ** 2] for i in range(200)]
    test_data_1d = [[(i + 1) ** 2] for i in range(10)]
    test_data_3d = [[i, (i + 1) ** 2, (i + 2) ** 2] for i in range(10)]
    sample_points = [[i, (i + 1) ** 2] for i in range(8)]
    sample_points_large = [[i, (i + 1) ** 2] for i in range(100)]
    sample_points_1d = [[(i + 1) ** 2] for i in range(8)]
    sample_points_3d = [[i, (i + 1) ** 2, (i + 2) ** 2] for i in range(8)]

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__01(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)

        PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                         maximum_polynomial_order=5)
        assert PolyClass.max_polynomial_order == 5
        assert PolyClass.number_of_crossvalidations == 3  # Default number of cross-validations
        assert PolyClass.no_adaptive_samples == 4  # Default number of adaptive samples
        assert PolyClass.fraction_training == 0.75  # Default training split
        assert PolyClass.max_fraction_training_samples == 0.5  # Default fraction for the maximum number of training samples
        assert PolyClass.max_iter == 10  # Default maximum number of iterations
        assert PolyClass.solution_method == 'pyomo'  # Default solution_method
        assert PolyClass.multinomials == 1  # Default multinomials

    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    @pytest.mark.unit
    def test__init__02(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                         maximum_polynomial_order=3,
                                         number_of_crossvalidations=5,
                                         no_adaptive_samples=6,
                                         training_split=0.5,
                                         max_fraction_training_samples=0.4,
                                         max_iter=20,
                                         solution_method='MLe',
                                         multinomials=0)
        assert PolyClass.max_polynomial_order == 3
        assert PolyClass.number_of_crossvalidations == 5  # Default number of cross-validations
        assert PolyClass.no_adaptive_samples == 6  # Default number of adaptive samples
        assert PolyClass.fraction_training == 0.5  # Default training split
        assert PolyClass.max_fraction_training_samples == 0.4  # Default fraction for the maximum number of training samples
        assert PolyClass.max_iter == 20  # Default maximum number of iterations
        assert PolyClass.solution_method == 'mle'  # Default solution_method, doesn't matter lower / upper characters
        assert PolyClass.multinomials == 0  # Default multinomials

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [list])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__03(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(ValueError):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=5)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [list])
    def test__init__04(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(ValueError):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=5)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__05(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points_large)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=5)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__06(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data_3d)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=5)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__07(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points_3d)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=5)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__08(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data_1d)
        regression_data_input = array_type2(self.sample_points_1d)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=5)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__09(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.warns(Warning):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=5,
                                             number_of_crossvalidations=11)
            assert PolyClass.number_of_crossvalidations == 11  # Default number of cross-validations

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__10(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=1.2)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__11(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data_large)
        regression_data_input = array_type2(self.sample_points_large)
        with pytest.warns(Warning):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=11)
            assert PolyClass.max_polynomial_order == 10

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__12(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=5,
                                             training_split=1)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__13(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=5,
                                             training_split=-1)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__14(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=5,
                                             max_fraction_training_samples=1.2)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__15(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=5,
                                             max_fraction_training_samples=-1.2)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__16(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        PolyClass = PolynomialRegression(regression_data_input, regression_data_input,
                                         maximum_polynomial_order=5,
                                         max_iter=100)
        assert PolyClass.max_iter == 0

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__17(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                         maximum_polynomial_order=5,
                                         no_adaptive_samples=0,
                                         max_iter=100)
        assert PolyClass.max_iter == 0

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__18(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=5,
                                             number_of_crossvalidations=1.2)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__19(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=5,
                                             no_adaptive_samples=4.2)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__20(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=5,
                                             max_iter=4.2)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__21(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=15)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__22(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=5,
                                             solution_method=1)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__23(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=5,
                                             solution_method='idaes')

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__24(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=5,
                                             multinomials=3)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__25(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=-2)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__26(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=3,
                                             number_of_crossvalidations=-3)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__27(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=3,
                                             no_adaptive_samples=-3)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__28(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=3,
                                             max_iter=-3)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__29(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=3,
                                             overwrite=1)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__30(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=3,
                                             fname='solution.pkl')

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__31(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        with pytest.raises(Exception):
            PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                             maximum_polynomial_order=3,
                                             fname=1)

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__32(self, array_type1, array_type2):
        file_name = 'sol_check.pickle'
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)

        PolyClass1 = PolynomialRegression(original_data_input, regression_data_input,
                                          maximum_polynomial_order=3,
                                          fname=file_name,
                                          overwrite=True)
        PolyClass1.get_feature_vector()
        results = PolyClass1.polynomial_regression_fitting()
        PolyClass2 = PolynomialRegression(original_data_input, regression_data_input,
                                          maximum_polynomial_order=3,
                                          fname=file_name,
                                          overwrite=True)
        assert PolyClass1.filename == PolyClass2.filename

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__33(self, array_type1, array_type2):
        file_name1 = 'sol_check1.pickle'
        file_name2 = 'sol_check2.pickle'
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)

        PolyClass1 = PolynomialRegression(original_data_input, regression_data_input,
                                          maximum_polynomial_order=3,
                                          fname=file_name1,
                                          overwrite=True)
        PolyClass1.get_feature_vector()
        results = PolyClass1.polynomial_regression_fitting()
        PolyClass2 = PolynomialRegression(original_data_input, regression_data_input,
                                          maximum_polynomial_order=3,
                                          fname=file_name2,
                                          overwrite=True)
        assert PolyClass1.filename == file_name1
        assert PolyClass2.filename == file_name2

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test__init__34(self, array_type1, array_type2):
        file_name = 'sol_check.pickle'
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        PolyClass1 = PolynomialRegression(original_data_input, regression_data_input, fname=file_name,
                                          maximum_polynomial_order=3, overwrite=True)
        PolyClass1.get_feature_vector()
        results = PolyClass1.polynomial_regression_fitting()
        PolygClass2 = PolynomialRegression(original_data_input, regression_data_input, fname=file_name,
                                           maximum_polynomial_order=3, overwrite=True)
        assert PolyClass1.filename == PolygClass2.filename

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test_training_test_data_creation_01(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)

        PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                         maximum_polynomial_order=5,
                                         training_split=0.01)
        with pytest.raises(Exception):
            training_data, cross_val_data = PolyClass.training_test_data_creation()

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array, pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array, pd.DataFrame])
    def test_training_test_data_creation_02(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)

        PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                         maximum_polynomial_order=5,
                                         training_split=0.99)
        with pytest.raises(Exception):
            training_data, cross_val_data = PolyClass.training_test_data_creation()

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_training_test_data_creation_03(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)

        PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                         maximum_polynomial_order=5)

        training_data, cross_val_data = PolyClass.training_test_data_creation()

        expected_training_size = int(np.around(PolyClass.number_of_samples * PolyClass.fraction_training))
        expected_test_size = PolyClass.regression_data.shape[0] - expected_training_size

        assert len(training_data) == PolyClass.number_of_crossvalidations
        assert len(cross_val_data) == PolyClass.number_of_crossvalidations

        for i in range(1, PolyClass.number_of_crossvalidations + 1):
            assert training_data["training_set_" + str(i)].shape[0] == expected_training_size
            assert cross_val_data["test_set_" + str(i)].shape[0] == expected_test_size

            concat_01 = (
                np.concatenate((training_data["training_set_" + str(i)], cross_val_data["test_set_" + str(i)]), axis=0))
            sample_data_sorted = regression_data_input[
                np.lexsort((regression_data_input[:, 2], regression_data_input[:, 1], regression_data_input[:, 0]))]
            concat_01_sorted = concat_01[np.lexsort((concat_01[:, 2], concat_01[:, 1], concat_01[:, 0]))]
            np.testing.assert_equal(sample_data_sorted, concat_01_sorted)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_training_test_data_creation_04(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                         maximum_polynomial_order=3,
                                         number_of_crossvalidations=5,
                                         no_adaptive_samples=6,
                                         training_split=0.5,
                                         max_fraction_training_samples=0.4,
                                         max_iter=20,
                                         solution_method='MLe',
                                         multinomials=0)
        training_data, cross_val_data = PolyClass.training_test_data_creation()
        expected_training_size = int(np.around(PolyClass.number_of_samples * PolyClass.fraction_training))
        expected_test_size = PolyClass.regression_data.shape[0] - expected_training_size
        assert len(training_data) == PolyClass.number_of_crossvalidations
        assert len(cross_val_data) == PolyClass.number_of_crossvalidations
        for i in range(1, PolyClass.number_of_crossvalidations + 1):
            assert training_data["training_set_" + str(i)].shape[0] == expected_training_size
            assert cross_val_data["test_set_" + str(i)].shape[0] == expected_test_size
            concat_01 = (
                np.concatenate((training_data["training_set_" + str(i)], cross_val_data["test_set_" + str(i)]), axis=0))
            sample_data_sorted = regression_data_input[
                np.lexsort((regression_data_input[:, 2], regression_data_input[:, 1], regression_data_input[:, 0]))]
            concat_01_sorted = concat_01[np.lexsort((concat_01[:, 2], concat_01[:, 1], concat_01[:, 0]))]
            np.testing.assert_equal(sample_data_sorted, concat_01_sorted)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_training_test_data_creation_05(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                         maximum_polynomial_order=5)
        additional_data_input = np.array(
            [[i ** 2, ((i + 1) * 2) + ((j + 1) * 2), j ** 4, ((i + 1) * 2) + ((j + 1) ** 2)] for i in range(5) for j in
             range(5)])
        training_data, cross_val_data = PolyClass.training_test_data_creation(additional_features=additional_data_input)
        expected_training_size = int(np.around(PolyClass.number_of_samples * PolyClass.fraction_training))
        expected_test_size = PolyClass.regression_data.shape[0] - expected_training_size
        assert len(training_data) == PolyClass.number_of_crossvalidations * 2
        assert len(cross_val_data) == PolyClass.number_of_crossvalidations * 2
        for i in range(1, PolyClass.number_of_crossvalidations + 1):
            assert training_data["training_set_" + str(i)].shape[0] == expected_training_size
            assert training_data["training_extras_" + str(i)].shape[0] == expected_training_size
            assert cross_val_data["test_set_" + str(i)].shape[0] == expected_test_size
            assert cross_val_data["test_extras_" + str(i)].shape[0] == expected_test_size
            concat_01 = (
                np.concatenate((training_data["training_set_" + str(i)], cross_val_data["test_set_" + str(i)]), axis=0))
            sample_data_sorted = regression_data_input[
                np.lexsort((regression_data_input[:, 2], regression_data_input[:, 1], regression_data_input[:, 0]))]
            concat_01_sorted = concat_01[np.lexsort((concat_01[:, 2], concat_01[:, 1], concat_01[:, 0]))]
            np.testing.assert_equal(sample_data_sorted, concat_01_sorted)
            concat_02 = (
                np.concatenate((training_data["training_extras_" + str(i)], cross_val_data["test_extras_" + str(i)]),
                               axis=0))
            additional_data_sorted = additional_data_input[
                np.lexsort((additional_data_input[:, 3], additional_data_input[:, 2], additional_data_input[:, 1],
                            additional_data_input[:, 0]))]
            concat_02_sorted = concat_02[
                np.lexsort((concat_02[:, 3], concat_02[:, 2], concat_02[:, 1], concat_02[:, 0]))]
            np.testing.assert_equal(additional_data_sorted, concat_02_sorted)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_polygeneration_01(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        x_input_train_data = regression_data_input[:, :-1]
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=1)
        poly_degree = 1
        output_1 = data_feed.polygeneration(poly_degree, data_feed.multinomials, x_input_train_data)
        expected_output_nr = x_input_train_data.shape[0]
        expected_output_nc = 4  # New number of features should be = 2 * max_polynomial_order + 2 for two input features
        expected_output = np.zeros((expected_output_nr, expected_output_nc))
        expected_output[:, 0] = 1
        expected_output[:, 1] = x_input_train_data[:, 0]
        expected_output[:, 2] = x_input_train_data[:, 1]
        expected_output[:, 3] = x_input_train_data[:, 0] * x_input_train_data[:, 1]
        np.testing.assert_equal(output_1, expected_output)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_polygeneration_02(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        x_input_train_data = regression_data_input[:, :-1]
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=2)
        poly_degree = 2
        output_1 = data_feed.polygeneration(poly_degree, data_feed.multinomials, x_input_train_data)
        expected_output_nr = x_input_train_data.shape[0]
        expected_output_nc = 6  # New number of features should be = 2 * max_polynomial_order + 2 for two input features
        expected_output = np.zeros((expected_output_nr, expected_output_nc))
        expected_output[:, 0] = 1
        expected_output[:, 1] = x_input_train_data[:, 0]
        expected_output[:, 2] = x_input_train_data[:, 1]
        expected_output[:, 3] = x_input_train_data[:, 0] ** 2
        expected_output[:, 4] = x_input_train_data[:, 1] ** 2
        expected_output[:, 5] = x_input_train_data[:, 0] * x_input_train_data[:, 1]
        np.testing.assert_equal(output_1, expected_output)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_polygeneration_03(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        x_input_train_data = regression_data_input[:, :-1]
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=10)
        poly_degree = 10
        output_1 = data_feed.polygeneration(poly_degree, data_feed.multinomials, x_input_train_data)
        expected_output_nr = x_input_train_data.shape[0]
        expected_output_nc = 22  # New number of features should be = 2 * max_polynomial_order + 2 for two input features
        expected_output = np.zeros((expected_output_nr, expected_output_nc))
        expected_output[:, 0] = 1
        expected_output[:, 1] = x_input_train_data[:, 0]
        expected_output[:, 2] = x_input_train_data[:, 1]
        expected_output[:, 3] = x_input_train_data[:, 0] ** 2
        expected_output[:, 4] = x_input_train_data[:, 1] ** 2
        expected_output[:, 5] = x_input_train_data[:, 0] ** 3
        expected_output[:, 6] = x_input_train_data[:, 1] ** 3
        expected_output[:, 7] = x_input_train_data[:, 0] ** 4
        expected_output[:, 8] = x_input_train_data[:, 1] ** 4
        expected_output[:, 9] = x_input_train_data[:, 0] ** 5
        expected_output[:, 10] = x_input_train_data[:, 1] ** 5
        expected_output[:, 11] = x_input_train_data[:, 0] ** 6
        expected_output[:, 12] = x_input_train_data[:, 1] ** 6
        expected_output[:, 13] = x_input_train_data[:, 0] ** 7
        expected_output[:, 14] = x_input_train_data[:, 1] ** 7
        expected_output[:, 15] = x_input_train_data[:, 0] ** 8
        expected_output[:, 16] = x_input_train_data[:, 1] ** 8
        expected_output[:, 17] = x_input_train_data[:, 0] ** 9
        expected_output[:, 18] = x_input_train_data[:, 1] ** 9
        expected_output[:, 19] = x_input_train_data[:, 0] ** 10
        expected_output[:, 20] = x_input_train_data[:, 1] ** 10
        expected_output[:, 21] = x_input_train_data[:, 0] * x_input_train_data[:, 1]
        np.testing.assert_equal(output_1, expected_output)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_polygeneration_04(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        x_input_train_data = regression_data_input[:, :-1]
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=10,
                                         multinomials=0)
        poly_degree = 10
        output_1 = data_feed.polygeneration(poly_degree, data_feed.multinomials, x_input_train_data)
        expected_output_nr = x_input_train_data.shape[0]
        expected_output_nc = 21  # New number of features should be = 2 * max_polynomial_order + 2 for two input features
        expected_output = np.zeros((expected_output_nr, expected_output_nc))
        expected_output[:, 0] = 1
        expected_output[:, 1] = x_input_train_data[:, 0]
        expected_output[:, 2] = x_input_train_data[:, 1]
        expected_output[:, 3] = x_input_train_data[:, 0] ** 2
        expected_output[:, 4] = x_input_train_data[:, 1] ** 2
        expected_output[:, 5] = x_input_train_data[:, 0] ** 3
        expected_output[:, 6] = x_input_train_data[:, 1] ** 3
        expected_output[:, 7] = x_input_train_data[:, 0] ** 4
        expected_output[:, 8] = x_input_train_data[:, 1] ** 4
        expected_output[:, 9] = x_input_train_data[:, 0] ** 5
        expected_output[:, 10] = x_input_train_data[:, 1] ** 5
        expected_output[:, 11] = x_input_train_data[:, 0] ** 6
        expected_output[:, 12] = x_input_train_data[:, 1] ** 6
        expected_output[:, 13] = x_input_train_data[:, 0] ** 7
        expected_output[:, 14] = x_input_train_data[:, 1] ** 7
        expected_output[:, 15] = x_input_train_data[:, 0] ** 8
        expected_output[:, 16] = x_input_train_data[:, 1] ** 8
        expected_output[:, 17] = x_input_train_data[:, 0] ** 9
        expected_output[:, 18] = x_input_train_data[:, 1] ** 9
        expected_output[:, 19] = x_input_train_data[:, 0] ** 10
        expected_output[:, 20] = x_input_train_data[:, 1] ** 10
        np.testing.assert_equal(output_1, expected_output)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_polygeneration_05(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        x_input_train_data = regression_data_input[:, :-1]
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=1)
        poly_degree = 1
        additional_term = np.sqrt(x_input_train_data)
        output_1 = data_feed.polygeneration(poly_degree, data_feed.multinomials, x_input_train_data, additional_term)
        expected_output_nr = x_input_train_data.shape[0]
        expected_output_nc = 6  # New number of features should be = 2 * max_polynomial_order + 4
        expected_output = np.zeros((expected_output_nr, expected_output_nc))
        expected_output[:, 0] = 1
        expected_output[:, 1] = x_input_train_data[:, 0]
        expected_output[:, 2] = x_input_train_data[:, 1]
        expected_output[:, 3] = x_input_train_data[:, 0] * x_input_train_data[:, 1]
        expected_output[:, 4] = additional_term[:, 0]
        expected_output[:, 5] = additional_term[:, 1]
        np.testing.assert_equal(output_1, expected_output)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_cost_function_01(self, array_type):
        regression_data_input = array_type(self.training_data)
        x = regression_data_input[:, :-1]
        y = regression_data_input[:, -1]
        x_data_nr = x.shape[0]
        x_data_nc = 6
        x_vector = np.zeros((x_data_nr, x_data_nc))
        x_vector[:, 0] = 1
        x_vector[:, 1] = x[:, 0]
        x_vector[:, 2] = x[:, 1]
        x_vector[:, 3] = x[:, 0] ** 2
        x_vector[:, 4] = x[:, 1] ** 2
        x_vector[:, 5] = x[:, 0] * x[:, 1]
        theta = np.zeros((x_data_nc, 1))
        expected_value = 6613.875  # Calculated externally as sum(y^2) / 2m
        output_1 = PolynomialRegression.cost_function(theta, x_vector, y, reg_parameter=0)
        assert output_1 == expected_value

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_cost_function_02(self, array_type):
        regression_data_input = array_type(self.training_data)
        x = regression_data_input[:, :-1]
        y = regression_data_input[:, -1]
        x_data_nr = x.shape[0]
        x_data_nc = 6
        x_vector = np.zeros((x_data_nr, x_data_nc))
        x_vector[:, 0] = 1
        x_vector[:, 1] = x[:, 0]
        x_vector[:, 2] = x[:, 1]
        x_vector[:, 3] = x[:, 0] ** 2
        x_vector[:, 4] = x[:, 1] ** 2
        x_vector[:, 5] = x[:, 0] * x[:, 1]
        theta = np.array([[4.5], [3], [3], [1], [1], [0]])  # coefficients in (x1 + 1.5)^2 + (x2 + 1.5) ^ 2
        expected_value = 90.625  # Calculated externally as sum(dy^2) / 2m
        output_1 = PolynomialRegression.cost_function(theta, x_vector, y, reg_parameter=0)
        assert output_1 == expected_value

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_cost_function_03(self, array_type):
        regression_data_input = array_type(self.training_data)
        x = regression_data_input[:, :-1]
        y = regression_data_input[:, -1]
        x_data_nr = x.shape[0]
        x_data_nc = 6
        x_vector = np.zeros((x_data_nr, x_data_nc))
        x_vector[:, 0] = 1
        x_vector[:, 1] = x[:, 0]
        x_vector[:, 2] = x[:, 1]
        x_vector[:, 3] = x[:, 0] ** 2
        x_vector[:, 4] = x[:, 1] ** 2
        x_vector[:, 5] = x[:, 0] * x[:, 1]
        theta = np.array([[2], [2], [2], [1], [1], [0]])  # Actual coefficients in (x1 + 1)^2 + (x2 + 1) ^ 2
        expected_value = 0  # Value should return zero for exact solution
        output_1 = PolynomialRegression.cost_function(theta, x_vector, y, reg_parameter=0)
        assert output_1 == expected_value

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_gradient_function_01(self, array_type):
        regression_data_input = array_type(self.training_data)
        x = regression_data_input[:, :-1]
        y = regression_data_input[:, -1]
        x_data_nr = x.shape[0]
        x_data_nc = 6
        x_vector = np.zeros((x_data_nr, x_data_nc))
        x_vector[:, 0] = 1
        x_vector[:, 1] = x[:, 0]
        x_vector[:, 2] = x[:, 1]
        x_vector[:, 3] = x[:, 0] ** 2
        x_vector[:, 4] = x[:, 1] ** 2
        x_vector[:, 5] = x[:, 0] * x[:, 1]
        theta = np.zeros((x_data_nc,))
        expected_value = np.array(
            [[-97], [-635], [-635], [-5246.875], [-5246.875], [-3925]])  # Calculated externally: see Excel sheet
        expected_value = expected_value.reshape(expected_value.shape[0], )
        output_1 = PolynomialRegression.gradient_function(theta, x_vector, y, reg_parameter=0)
        np.testing.assert_equal(output_1, expected_value)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_gradient_function_02(self, array_type):
        regression_data_input = array_type(self.training_data)
        x = regression_data_input[:, :-1]
        y = regression_data_input[:, -1]
        x_data_nr = x.shape[0]
        x_data_nc = 6
        x_vector = np.zeros((x_data_nr, x_data_nc))
        x_vector[:, 0] = 1
        x_vector[:, 1] = x[:, 0]
        x_vector[:, 2] = x[:, 1]
        x_vector[:, 3] = x[:, 0] ** 2
        x_vector[:, 4] = x[:, 1] ** 2
        x_vector[:, 5] = x[:, 0] * x[:, 1]
        theta = np.array([[4.5], [3], [3], [1], [1], [0]])  # coefficients in (x1 + 1.5)^2 + (x2 + 1.5) ^ 2
        theta = theta.reshape(theta.shape[0], )
        expected_value = np.array(
            [[12.5], [75], [75], [593.75], [593.75], [437.5]])  # Calculated externally: see Excel sheet
        expected_value = expected_value.reshape(expected_value.shape[0], )
        output_1 = PolynomialRegression.gradient_function(theta, x_vector, y, reg_parameter=0)
        np.testing.assert_equal(output_1, expected_value)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_gradient_function_03(self, array_type):
        regression_data_input = array_type(self.training_data)
        x = regression_data_input[:, :-1]
        y = regression_data_input[:, -1]
        x_data_nr = x.shape[0]
        x_data_nc = 6
        x_vector = np.zeros((x_data_nr, x_data_nc))
        x_vector[:, 0] = 1
        x_vector[:, 1] = x[:, 0]
        x_vector[:, 2] = x[:, 1]
        x_vector[:, 3] = x[:, 0] ** 2
        x_vector[:, 4] = x[:, 1] ** 2
        x_vector[:, 5] = x[:, 0] * x[:, 1]
        theta = np.array([[2], [2], [2], [1], [1], [0]])  # Actual coefficients in (x1 + 1)^2 + (x2 + 1) ^ 2
        theta = theta.reshape(theta.shape[0], )
        expected_value = np.array([[0], [0], [0], [0], [0], [0]])  # Calculated externally: see Excel sheet
        expected_value = expected_value.reshape(expected_value.shape[0], )
        output_1 = PolynomialRegression.gradient_function(theta, x_vector, y, reg_parameter=0)
        np.testing.assert_equal(output_1, expected_value)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_bfgs_parameter_optimization_01(self, array_type):
        original_data_input = array_type(self.test_data)
        # Create x vector for ax2 + bx + c: x data supplied in x_vector
        input_array = np.array([[0, 1], [1, 4], [2, 9], [3, 16], [4, 25], [5, 36], [6, 49], [7, 64], [8, 81], [9, 100]])
        x = input_array[:, 0]
        y = input_array[:, 1]
        x_vector = np.zeros((x.shape[0], 3))
        x_vector[:, 0] = x[:, ] ** 2
        x_vector[:, 1] = x[:, ]
        x_vector[:, 2] = 1
        expected_value = np.array([[1.], [2.], [1.]]).reshape(3, )
        data_feed = PolynomialRegression(original_data_input, input_array, maximum_polynomial_order=5,
                                         solution_method='bfgs')
        output_1 = data_feed.bfgs_parameter_optimization(x_vector, y)
        assert data_feed.solution_method == 'bfgs'
        np.testing.assert_array_equal(expected_value, np.round(output_1, 4))

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_bfgs_parameter_optimization_02(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        # Create x vector for ax2 + bx + c: x data supplied in x_vector
        x = regression_data_input[:, : -1]
        y = regression_data_input[:, -1]
        x_vector = np.zeros((x.shape[0], 6))
        x_vector[:, 0] = x[:, 0] ** 2
        x_vector[:, 1] = x[:, 1] ** 2
        x_vector[:, 2] = x[:, 0]
        x_vector[:, 3] = x[:, 1]
        x_vector[:, 4] = x[:, 1] * x[:, 0]
        x_vector[:, 5] = 1
        expected_value = np.array([[1.], [1.], [2.], [2.], [0.], [2.]]).reshape(6, )
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=4,
                                         solution_method='bfgs')
        output_1 = data_feed.bfgs_parameter_optimization(x_vector, y)
        assert data_feed.solution_method == 'bfgs'
        np.testing.assert_array_equal(expected_value, np.round(output_1, 4))

    @pytest.mark.unit
    @staticmethod
    def test_mle_estimate_01(self):
        # Create x vector for ax2 + bx + c: x data supplied in x_vector
        input_array = np.array([[0, 1], [1, 4], [2, 9], [3, 16], [4, 25], [5, 36], [6, 49], [7, 64], [8, 81], [9, 100]])
        x = input_array[:, 0]
        y = input_array[:, 1]
        x_vector = np.zeros((x.shape[0], 3))
        x_vector[:, 0] = x[:, ] ** 2
        x_vector[:, 1] = x[:, ]
        x_vector[:, 2] = 1
        expected_value = np.array([[1.], [2.], [1.]]).reshape(3, )
        output_1 = PolynomialRegression.MLE_estimate(x_vector, y)
        np.testing.assert_array_equal(expected_value, np.round(output_1, 4))

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_mle_estimate_02(self, array_type):
        regression_data_input = array_type(self.training_data)
        x = regression_data_input[:, : -1]
        y = regression_data_input[:, -1]
        x_vector = np.zeros((x.shape[0], 6))
        x_vector[:, 0] = x[:, 0] ** 2
        x_vector[:, 1] = x[:, 1] ** 2
        x_vector[:, 2] = x[:, 0]
        x_vector[:, 3] = x[:, 1]
        x_vector[:, 4] = x[:, 1] * x[:, 0]
        x_vector[:, 5] = 1
        expected_value = np.array([[1.], [1.], [2.], [2.], [0.], [2.]]).reshape(6, )
        output_1 = PolynomialRegression.MLE_estimate(x_vector, y)
        np.testing.assert_array_equal(expected_value, np.round(output_1, 4))

    @pytest.mark.unit
    def test_pyomo_optimization_01(self):
        x_vector = np.array([[i ** 2, i, 1] for i in range(10)])
        y = np.array([[i ** 2] for i in range(1, 11)])
        expected_value = np.array([[1.], [2.], [1.]])
        output_1 = PolynomialRegression.pyomo_optimization(x_vector, y)
        np.testing.assert_array_equal(expected_value, np.round(output_1, 4))

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_pyomo_optimization_02(self, array_type):
        regression_data_input = array_type(self.training_data)
        x = regression_data_input[:, : -1]
        y = regression_data_input[:, -1]
        x_vector = np.zeros((x.shape[0], 6))
        x_vector[:, 0] = x[:, 0] ** 2
        x_vector[:, 1] = x[:, 1] ** 2
        x_vector[:, 2] = x[:, 0]
        x_vector[:, 3] = x[:, 1]
        x_vector[:, 4] = x[:, 1] * x[:, 0]
        x_vector[:, 5] = 1
        expected_value = np.array([[1.], [1.], [2.], [2.], [0.], [2.]])
        output_1 = PolynomialRegression.pyomo_optimization(x_vector, y)
        np.testing.assert_array_equal(expected_value, np.round(output_1, 4))

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_cross_validation_error_calculation_01(self, array_type):
        regression_data_input = array_type(self.training_data)
        x = regression_data_input[:, :-1]
        y = regression_data_input[:, -1].reshape(regression_data_input.shape[0], 1)
        x_data_nr = x.shape[0]
        x_data_nc = 6
        x_vector = np.zeros((x_data_nr, x_data_nc))
        x_vector[:, 0] = 1
        x_vector[:, 1] = x[:, 0]
        x_vector[:, 2] = x[:, 1]
        x_vector[:, 3] = x[:, 0] ** 2
        x_vector[:, 4] = x[:, 1] ** 2
        x_vector[:, 5] = x[:, 0] * x[:, 1]
        theta = np.zeros((x_data_nc, 1))
        expected_value = 2 * 6613.875  # Calculated externally as sum(y^2) / m
        output_1 = PolynomialRegression.cross_validation_error_calculation(theta, x_vector, y)
        assert output_1 == expected_value

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_cross_validation_error_calculation_02(self, array_type):
        regression_data_input = array_type(self.training_data)
        x = regression_data_input[:, :-1]
        y = regression_data_input[:, -1].reshape(regression_data_input.shape[0], 1)
        x_data_nr = x.shape[0]
        x_data_nc = 6
        x_vector = np.zeros((x_data_nr, x_data_nc))
        x_vector[:, 0] = 1
        x_vector[:, 1] = x[:, 0]
        x_vector[:, 2] = x[:, 1]
        x_vector[:, 3] = x[:, 0] ** 2
        x_vector[:, 4] = x[:, 1] ** 2
        x_vector[:, 5] = x[:, 0] * x[:, 1]
        theta = np.array([[4.5], [3], [3], [1], [1], [0]])  # coefficients in (x1 + 1.5)^2 + (x2 + 1.5) ^ 2
        expected_value = 2 * 90.625  # Calculated externally as sum(dy^2) / 2m
        output_1 = PolynomialRegression.cross_validation_error_calculation(theta, x_vector, y)
        assert output_1 == expected_value

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_cross_validation_error_calculation_02(self, array_type):
        regression_data_input = array_type(self.training_data)
        x = regression_data_input[:, :-1]
        y = regression_data_input[:, -1].reshape(regression_data_input.shape[0], 1)
        x_data_nr = x.shape[0]
        x_data_nc = 6
        x_vector = np.zeros((x_data_nr, x_data_nc))
        x_vector[:, 0] = 1
        x_vector[:, 1] = x[:, 0]
        x_vector[:, 2] = x[:, 1]
        x_vector[:, 3] = x[:, 0] ** 2
        x_vector[:, 4] = x[:, 1] ** 2
        x_vector[:, 5] = x[:, 0] * x[:, 1]
        theta = np.array([[2], [2], [2], [1], [1], [0]])  # Actual coefficients in (x1 + 1)^2 + (x2 + 1) ^ 2
        expected_value = 2 * 0  # Value should return zero for exact solution
        output_1 = PolynomialRegression.cross_validation_error_calculation(theta, x_vector, y)
        assert output_1 == expected_value

    def mock_optimization(self, x, y):
        return 10 * np.ones((x.shape[1], 1))

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    @patch.object(PolynomialRegression, 'MLE_estimate', mock_optimization)
    def test_polyregression_01(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=5,
                                         solution_method='mle')
        poly_order = 2
        training_data = regression_data_input[0:20, :]
        test_data = regression_data_input[20:, :]
        expected_output = 10 * np.ones((6, 1))
        output_1, _, _ = data_feed.polyregression(poly_order, training_data, test_data)
        np.testing.assert_array_equal(expected_output, output_1)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    @patch.object(PolynomialRegression, 'bfgs_parameter_optimization', mock_optimization)
    def test_polyregression_02(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=5,
                                         solution_method='bfgs')
        poly_order = 2
        training_data = regression_data_input[0:20, :]
        test_data = regression_data_input[20:, :]
        expected_output = 10 * np.ones((6, 1))
        output_1, _, _ = data_feed.polyregression(poly_order, training_data, test_data)
        np.testing.assert_array_equal(expected_output, output_1)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    @patch.object(PolynomialRegression, 'pyomo_optimization', mock_optimization)
    def test_polyregression_03(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=5)
        poly_order = 2
        training_data = regression_data_input[0:20, :]
        test_data = regression_data_input[20:, :]
        expected_output = 10 * np.ones((6, 1))
        output_1, _, _ = data_feed.polyregression(poly_order, training_data, test_data)
        np.testing.assert_array_equal(expected_output, output_1)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_polyregression_04(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=5)
        poly_order = 10
        training_data = regression_data_input[0:20, :]
        test_data = regression_data_input[20:, :]
        expected_output = np.Inf
        output_1, output_2, output_3 = data_feed.polyregression(poly_order, training_data, test_data)
        np.testing.assert_array_equal(expected_output, output_1)
        np.testing.assert_array_equal(expected_output, output_2)
        np.testing.assert_array_equal(expected_output, output_3)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_surrogate_performance_01(self, array_type):
        original_data_input = array_type(self.test_data)
        input_array = np.array([[0, 1], [1, 4], [2, 9], [3, 16], [4, 25], [5, 36], [6, 49], [7, 64], [8, 81], [9, 100]])
        order_best = 2
        phi_best = np.array([[0.], [0.], [0.]])
        expected_value_1 = 38.5
        expected_value_2 = 2533.3
        expected_value_3 = -1.410256
        expected_value_4 = 0
        data_feed = PolynomialRegression(original_data_input, input_array, maximum_polynomial_order=5)
        _, output_1, output_2, output_3, output_4 = data_feed.surrogate_performance(phi_best, order_best)
        assert output_1 == expected_value_1
        assert output_2 == expected_value_2
        assert np.round(output_3, 4) == np.round(expected_value_3, 4)
        assert np.round(output_4, 4) == np.round(expected_value_4, 4)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_surrogate_performance_02(self, array_type):
        original_data_input = array_type(self.test_data)
        input_array = np.array([[0, 1], [1, 4], [2, 9], [3, 16], [4, 25], [5, 36], [6, 49], [7, 64], [8, 81], [9, 100]])
        order_best = 2
        phi_best = np.array([[1.], [2.], [1.]])
        expected_value_1 = 0
        expected_value_2 = 0
        expected_value_3 = 1
        expected_value_4 = 1
        data_feed = PolynomialRegression(original_data_input, input_array, maximum_polynomial_order=5)
        _, output_1, output_2, output_3, output_4 = data_feed.surrogate_performance(phi_best, order_best)
        assert output_1 == expected_value_1
        assert output_2 == expected_value_2
        assert np.round(output_3, 4) == np.round(expected_value_3, 4)
        assert np.round(output_4, 4) == np.round(expected_value_4, 4)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_surrogate_performance_03(self, array_type):
        original_data_input = array_type(self.test_data)
        input_array = np.array([[0, 1], [1, 4], [2, 9], [3, 16], [4, 25], [5, 36], [6, 49], [7, 64], [8, 81], [9, 100]])
        order_best = 2
        phi_best = np.array([[1.], [1.], [1.]])
        expected_value_1 = 4.5
        expected_value_2 = 28.5
        expected_value_3 = 0.972884259
        expected_value_4 = 0.9651369
        data_feed = PolynomialRegression(original_data_input, input_array, maximum_polynomial_order=5)
        _, output_1, output_2, output_3, output_4 = data_feed.surrogate_performance(phi_best, order_best)
        assert output_1 == expected_value_1
        assert output_2 == expected_value_2
        assert np.round(output_3, 4) == np.round(expected_value_3, 4)
        assert np.round(output_4, 4) == np.round(expected_value_4, 4)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_results_generation_01(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=3,
                                         multinomials=0)
        order = 1
        beta = np.array([[0], [0], [0]])
        expected_df = pd.Series()
        row_list = np.array([['k'], ['(x_1)^1'], ['(x_2)^1']])
        expected_df = expected_df.append(
            pd.Series({row_list[0, 0]: beta[0, 0], row_list[1, 0]: beta[1, 0], row_list[2, 0]: beta[2, 0]}))
        output_df = data_feed.results_generation(beta, order)
        assert output_df.index.to_list() == expected_df.index.to_list()
        assert expected_df.all() == output_df.all()

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_results_generation_02(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=3,
                                         multinomials=0)
        order = 3
        beta = np.array([[1], [0.3], [6], [500], [500000], [0.001], [50]])
        expected_df = pd.Series()
        row_list = np.array([['k'], ['(x_1)^1'], ['(x_2)^1'], ['(x_1)^2'], ['(x_2)^2'], ['(x_1)^3'], ['(x_2)^3']])
        expected_df = expected_df.append(pd.Series(
            {row_list[0, 0]: beta[0, 0], row_list[1, 0]: beta[1, 0], row_list[2, 0]: beta[2, 0],
             row_list[3, 0]: beta[3, 0], row_list[4, 0]: beta[4, 0], row_list[5, 0]: beta[5, 0],
             row_list[6, 0]: beta[6, 0]}))
        output_df = data_feed.results_generation(beta, order)
        assert output_df.index.to_list() == expected_df.index.to_list()
        assert expected_df.all() == output_df.all()

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_results_generation_03(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=3,
                                         multinomials=1)
        order = 2
        beta = np.array([[1], [0.3], [6], [500], [500000], [0.001]])
        expected_df = pd.Series()
        row_list = np.array([['k'], ['(x_1)^1'], ['(x_2)^1'], ['(x_1)^2'], ['(x_2)^2'], ['(x_1).(x_2)']])
        expected_df = expected_df.append(pd.Series(
            {row_list[0, 0]: beta[0, 0], row_list[1, 0]: beta[1, 0], row_list[2, 0]: beta[2, 0],
             row_list[3, 0]: beta[3, 0], row_list[4, 0]: beta[4, 0], row_list[5, 0]: beta[5, 0]}))
        output_df = data_feed.results_generation(beta, order)
        assert output_df.index.to_list() == expected_df.index.to_list()
        assert expected_df.all() == output_df.all()

    @pytest.mark.unit
    @patch("matplotlib.pyplot.show")
    def test_error_plotting(self, mock_show):
        mock_show.return_value = None
        # Generate typical data values for eaxch variable in order to test plot function
        plotting_data = np.array(
            [[1, 7, 23.1, 29.2, 0.01, 1300, -1.6, -1.1, 5], [2, 9, 0.0055, 0.0015, 0.006, 159, 0.34, 0.19, 10],
             [3, 6, 0.0007, 0.0009, 0.001, 2.3, 0.998, 0.994, 15], [4, 2, 0.0005, 0.0004, 0.0008, 0.02, 1.0, 1.0, 20]])
        ax1, ax2, ax3, ax4 = PolynomialRegression.error_plotting(plotting_data)
        expected_output_1 = np.array([[1, 23.1], [2, 0.0055], [3, 0.0007], [4, 0.0005]])
        expected_output_2 = np.array([[1, 29.2], [2, 0.0015], [3, 0.0009], [4, 0.0004]])
        expected_output_3 = np.array([[1, 0.01], [2, 0.006], [3, 0.001], [4, 0.0008]])
        expected_output_4 = np.array([[1, 1300], [2, 159], [3, 2.3], [4, 0.02]])
        expected_output_5 = np.array([[1, -1.6], [2, 0.34], [3, 0.998], [4, 1.0]])
        expected_output_6 = np.array([[1, -1.1], [2, 0.19], [3, 0.994], [4, 1.0]])
        # Compare the data input to the actual plotted data
        np.testing.assert_array_equal(ax1.lines[0].get_xydata(), expected_output_1)
        np.testing.assert_array_equal(ax1.lines[1].get_xydata(), expected_output_2)
        np.testing.assert_array_equal(ax2.lines[0].get_xydata(), expected_output_3)
        np.testing.assert_array_equal(ax3.lines[0].get_xydata(), expected_output_4)
        np.testing.assert_array_equal(ax4.lines[0].get_xydata(), expected_output_5)
        np.testing.assert_array_equal(ax4.lines[1].get_xydata(), expected_output_6)
        # Compare the expected subplot titles to the actual display
        assert ax1.title.get_text() == 'Training (green) vs Cross-validation error (red)'
        assert ax2.title.get_text() == 'MAE'
        assert ax3.title.get_text() == 'MSE'
        assert ax4.title.get_text() == 'R-squared (blue) and Adjusted R-squared (red)'
        # Compare the line/curve segments for each line plot to the actual input data
        np.testing.assert_array_equal(ax1.lines[0].get_path()._vertices, expected_output_1)
        np.testing.assert_array_equal(ax1.lines[1].get_path()._vertices, expected_output_2)
        np.testing.assert_array_equal(ax2.lines[0].get_path()._vertices, expected_output_3)
        np.testing.assert_array_equal(ax3.lines[0].get_path()._vertices, expected_output_4)
        np.testing.assert_array_equal(ax4.lines[0].get_path()._vertices, expected_output_5)
        np.testing.assert_array_equal(ax4.lines[1].get_path()._vertices, expected_output_6)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_user_defined_terms_01(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        num_cv = 3
        split = 0.75
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=2,
                                         training_split=split, number_of_crossvalidations=num_cv)
        additional_terms = [np.sin(regression_data_input[:, 0]), np.sin(regression_data_input[:, 1])]
        returned_features_array = data_feed.user_defined_terms(additional_terms)

        # Create two additional columns
        expected_array = np.zeros((regression_data_input.shape[0], 2))
        expected_array[:, 0] = np.sin(regression_data_input[:, 0])
        expected_array[:, 1] = np.sin(regression_data_input[:, 1])

        np.testing.assert_equal(expected_array, returned_features_array)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [pd.DataFrame])
    def test_user_defined_terms_02(self, array_type):
        original_data_input = array_type(self.full_data)
        num_cv = 3
        split = 0.75
        data_feed = PolynomialRegression(original_data_input, original_data_input, maximum_polynomial_order=2,
                                         training_split=split, number_of_crossvalidations=num_cv)
        additional_terms = [np.sin(original_data_input['x1']), np.sin(original_data_input['x2'])]
        returned_features_array = data_feed.user_defined_terms(additional_terms)

        # Create two additional columns
        expected_array = np.zeros((original_data_input.values.shape[0], 2))
        expected_array[:, 0] = np.sin(original_data_input.values[:, 0])
        expected_array[:, 1] = np.sin(original_data_input.values[:, 1])

        np.testing.assert_equal(expected_array, returned_features_array)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_user_defined_terms_03(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        num_cv = 3
        split = 0.75
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=2,
                                         training_split=split, number_of_crossvalidations=num_cv)
        additional_terms = [np.sin(original_data_input['x1']), np.sin(original_data_input['x2'])]
        with pytest.raises(Exception):
            data_feed.user_defined_terms(additional_terms)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_user_defined_terms_04(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        num_cv = 3
        split = 0.75
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=2,
                                         training_split=split, number_of_crossvalidations=num_cv)
        p = np.sin(regression_data_input[:, 0]).reshape(regression_data_input.shape[0],
                                                        1)  # 2-D array, should raise error
        additional_terms = [p]
        with pytest.raises(Exception):
            data_feed.user_defined_terms(additional_terms)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_user_defined_terms_05(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        num_cv = 3
        split = 0.75
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=2,
                                         training_split=split, number_of_crossvalidations=num_cv)
        p = np.sin(regression_data_input[:, 0]).reshape(regression_data_input.shape[0], 1)
        additional_terms = p  # Additional terms as array, not list
        with pytest.raises(ValueError):
            data_feed.user_defined_terms(additional_terms)

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_polynomial_regression_fitting_01(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=2)
        data_feed.get_feature_vector()
        results = data_feed.polynomial_regression_fitting()
        assert results.fit_status == 'ok'

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    @patch("matplotlib.pyplot.show")
    def test_polynomial_regression_fitting_02(self, mock_show, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        mock_show.return_value = None
        data_feed = PolynomialRegression(original_data_input, regression_data_input[:5], maximum_polynomial_order=1)
        data_feed.get_feature_vector()
        with pytest.warns(Warning):
            results = data_feed.polynomial_regression_fitting()
            assert results.fit_status == 'poor'

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_polynomial_regression_fitting_03(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=2)
        data_feed.get_feature_vector()
        results = data_feed.polynomial_regression_fitting()
        x_input_train_data = regression_data_input[:, :-1]
        assert results.fit_status == 'ok'

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_polynomial_regression_fitting_04(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=2)
        data_feed.get_feature_vector()
        additional_regression_features = [np.sin(regression_data_input[:, 0]), np.sin(regression_data_input[:, 1])]
        results = data_feed.polynomial_regression_fitting(additional_regression_features)
        results = data_feed.polynomial_regression_fitting()
        assert results.fit_status == 'ok'

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_get_feature_vector_01(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=3,
                                         multinomials=0)
        output = data_feed.get_feature_vector()
        expected_dict = {'x1': 0, 'x2': 0}
        assert expected_dict == output.extract_values()

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_get_feature_vector_02(self, array_type):
        regression_data_input = array_type(self.training_data)
        data_feed = PolynomialRegression(regression_data_input, regression_data_input, maximum_polynomial_order=3,
                                         multinomials=0)
        output = data_feed.get_feature_vector()
        expected_dict = {0: 0, 1: 0}
        assert expected_dict == output.extract_values()

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_set_additional_terms_01(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=3)
        data_feed.set_additional_terms('a')
        assert 'a' == data_feed.additional_term_expressions
        data_feed.set_additional_terms(1)
        assert 1 == data_feed.additional_term_expressions
        data_feed.set_additional_terms([1, 2])
        assert [1, 2] == data_feed.additional_term_expressions
        data_feed.set_additional_terms(np.array([1, 2]))
        np.testing.assert_equal(np.array([1, 2]), data_feed.additional_term_expressions)

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_poly_training_01(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=2)
        data_feed.get_feature_vector()
        data_feed.training()
        assert data_feed.fit_status == 'ok'

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type1", [pd.DataFrame])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_generate_expression(self, array_type1, array_type2):
        original_data_input = array_type1(self.full_data)
        regression_data_input = array_type2(self.training_data)
        data_feed = PolynomialRegression(original_data_input, regression_data_input, maximum_polynomial_order=2)

        p = data_feed.get_feature_vector()
        data_feed.training()

        lv = []
        for i in p.keys():
            lv.append(p[i])
        poly_expr = data_feed.generate_expression((lv))

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type1", [np.array])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_pickle_load01(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                         maximum_polynomial_order=3)
        PolyClass.get_feature_vector()
        PolyClass.training()
        PolyClass.pickle_load(PolyClass.filename)

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type1", [np.array])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_pickle_load02(self, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                         maximum_polynomial_order=3)
        with pytest.raises(Exception):
            PolyClass.pickle_load('file_not_existing.pickle')

    @pytest.mark.unit
    @patch("matplotlib.pyplot.show")
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type1", [np.array])
    @pytest.mark.parametrize("array_type2", [np.array])
    def test_parity_residual_plots(self, mock_show, array_type1, array_type2):
        original_data_input = array_type1(self.test_data)
        regression_data_input = array_type2(self.sample_points)
        PolyClass = PolynomialRegression(original_data_input, regression_data_input,
                                         maximum_polynomial_order=3)
        PolyClass.get_feature_vector()
        PolyClass.training()

        PolyClass.parity_residual_plots()

    @pytest.fixture(scope='module')
    @pytest.mark.unit
    def test_confint_regression_01(self):
        # Create x vector for ax2 + bx + c: x data supplied in x_vector
        input_array = np.array([[0, 1], [1, 4], [2, 9], [3, 16], [4, 25], [5, 36], [6, 49], [7, 64], [8, 81], [9, 100]])
        expected_stderror = np.array([4.541936, 0.850783])
        data_feed = PolynomialRegression(input_array, input_array, maximum_polynomial_order=1, overwrite=True)
        p = data_feed.get_feature_vector()
        res = data_feed.training()
        opt_wts = res.optimal_weights_array.reshape(expected_stderror.shape)
        output = data_feed.confint_regression(confidence=0.99)
        tval = 3.2498355440153697
        np.testing.assert_allclose(output['Std. error'].values, expected_stderror, atol=1e-3)
        np.testing.assert_allclose(output['Conf. int. lower'].values, opt_wts - tval * expected_stderror, atol=1e-3)
        np.testing.assert_allclose(output['Conf. int. upper'].values, opt_wts + tval * expected_stderror, atol=1e-3)

    @pytest.fixture(scope='module')
    @pytest.mark.unit
    def test_confint_regression_02(self):
        # Create x vector for ax2 + bx + c: x data supplied in x_vector
        input_array = np.array([[0, 1], [1, 4], [2, 9], [3, 16], [4, 25], [5, 36], [6, 49], [7, 64], [8, 81], [9, 100]])
        expected_stderror = np.array([4.541936, 0.850783])
        data_feed = PolynomialRegression(input_array, input_array, maximum_polynomial_order=1, overwrite=True)
        p = data_feed.get_feature_vector()
        res = data_feed.training()
        opt_wts = res.optimal_weights_array.reshape(expected_stderror.shape)
        output = data_feed.confint_regression(confidence=0.9)
        tval = 1.8331129326536335
        np.testing.assert_allclose(output['Std. error'].values, expected_stderror, atol=1e-3)
        np.testing.assert_allclose(output['Conf. int. lower'].values, opt_wts - tval * expected_stderror, atol=1e-3)
        np.testing.assert_allclose(output['Conf. int. upper'].values, opt_wts + tval * expected_stderror, atol=1e-3)

    @pytest.fixture(scope='module')
    @pytest.mark.unit
    def test_confint_regression_03(self):
        # Create x vector for ax2 + bx + c: x data supplied in x_vector
        input_array = np.array([[0, 1], [1, 4], [2, 9], [3, 16], [4, 25], [5, 36], [6, 49], [7, 64], [8, 81], [9, 100]])
        expected_stderror = np.array([0, 0, 0])
        data_feed = PolynomialRegression(input_array, input_array, maximum_polynomial_order=2, overwrite=True)
        p = data_feed.get_feature_vector()
        res = data_feed.training()
        opt_wts = res.optimal_weights_array.reshape(expected_stderror.shape)
        output = data_feed.confint_regression(confidence=0.99)
        tval = 3.2498355440153697
        np.testing.assert_allclose(output['Std. error'].values, expected_stderror, atol=1e-3)
        np.testing.assert_allclose(output['Conf. int. lower'].values, opt_wts - tval * expected_stderror, atol=1e-3)
        np.testing.assert_allclose(output['Conf. int. upper'].values, opt_wts + tval * expected_stderror, atol=1e-3)


if __name__ == '__main__':
    pytest.main()




